def arm_multiply(cpu: CPU, instr: int):
    """Execute a Multiply instruction"""
    rd = get_bits(instr, 16, 19)
    rn = get_bits(instr, 12, 15)
    rs = get_bits(instr, 8, 11)
    rm = get_bits(instr, 0, 3)
    accumulate = get_bit(instr, 21)
    set_cond_codes = get_bit(instr, 20)
    arm_multiply_idle(cpu, cpu.regs.get(rs), True)
    mask = 4294967295
    if accumulate:
        cpu.scheduler.idle(1)
        cpu.regs.set(rd, cpu.regs.get(rm) * cpu.regs.get(rs) + cpu.regs.get(rn) & mask)
    else:
        cpu.regs.set(rd, cpu.regs.get(rm) * cpu.regs.get(rs) & mask)
    if set_cond_codes:
        cpu.regs.cpsr.set_sign_flag(sign_32(cpu.regs.get(rd)))
        cpu.regs.cpsr.set_zero_flag(cpu.regs.get(rd) == 0)
    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.NON_SEQUENTIAL

def arm_multiply_long(cpu: CPU, instr: int):
    """Execute a Multiply Long instruction"""
    rd_hi = get_bits(instr, 16, 19)
    rd_lo = get_bits(instr, 12, 15)
    rs = get_bits(instr, 8, 11)
    rm = get_bits(instr, 0, 3)
    rd_hi_val = cpu.regs.get(rd_hi)
    rd_lo_val = cpu.regs.get(rd_lo)
    rs_val = cpu.regs.get(rs)
    rm_val = cpu.regs.get(rm)
    opcode = get_bits(instr, 21, 22)
    set_cond_codes = get_bit(instr, 20)
    cpu.scheduler.idle(1)
    if opcode == MultiplyLongOpcode.UMULL:
        arm_multiply_idle(cpu, rs_val, False)
        result = rm_val * rs_val
    elif opcode == MultiplyLongOpcode.UMLAL:
        arm_multiply_idle(cpu, rs_val, False)
        cpu.scheduler.idle(1)
        result = rd_hi_val << 32 | rd_lo_val
        result += rm_val * rs_val
    elif opcode == MultiplyLongOpcode.SMULL:
        arm_multiply_idle(cpu, rs_val, True)
        result = extend_sign_32(rm_val) * extend_sign_32(rs_val)
    elif opcode == MultiplyLongOpcode.SMLAL:
        arm_multiply_idle(cpu, rs_val, True)
        cpu.scheduler.idle(1)
        result = rd_hi_val << 32 | rd_lo_val
        result += extend_sign_32(rm_val) * extend_sign_32(rs_val)
    mask = 4294967295
    cpu.regs.set(rd_lo, result & mask)
    cpu.regs.set(rd_hi, result >> 32 & mask)
    if set_cond_codes:
        cpu.regs.cpsr.set_sign_flag(sign_32(cpu.regs.get(rd_hi)))
        cpu.regs.cpsr.set_zero_flag(cpu.regs.get(rd_hi) == 0 and cpu.regs.get(rd_lo) == 0)
    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.NON_SEQUENTIAL

def arm_multiply_idle(cpu: CPU, rs: int, signed: bool):
    """
    Idle for the number of cycles it takes to execute a multiply instruction.
    The number of cycles depends on how many MSBs of the Rs operand are
    "all 0" (if unsigned) or "all 0 or all 1" (if signed).
    """
    cycles = 1
    mask = 4294967040
    while mask != 0:
        rs &= mask
        if rs == 0 or (signed and rs == mask):
            break
        cycles += 1
        mask <<= 8
    cpu.scheduler.idle(cycles)


cimport cython

from libc.stdint cimport uint32_t
from cpython.array cimport array

from pyboy_advance.cpu.constants cimport CPUMode, CPUState, BankIndex
from pyboy_advance.utils cimport get_bit, set_bit


@cython.collection_type("sequence")
cdef class Registers:
    cdef int SP, LR, PC
    cdef int BANKED_GPR_RANGE_START
    cdef int BANKED_GPR_RANGE_END
    cdef int BANKED_GPR_RANGE_LEN

    cdef uint32_t[:] regs
    cdef ProgramStatusRegister cpsr
    cdef ProgramStatusRegister spsr
    cdef uint32_t[:] banked_old_gpr
    cdef uint32_t[:] banked_fiq_gpr
    cdef uint32_t[:] banked_sp
    cdef uint32_t[:] banked_lr
    cdef ProgramStatusRegister banked_spsr_user
    cdef ProgramStatusRegister banked_spsr_fiq
    cdef ProgramStatusRegister banked_spsr_irq
    cdef ProgramStatusRegister banked_spsr_swi
    cdef ProgramStatusRegister banked_spsr_abort
    cdef ProgramStatusRegister banked_spsr_undefined

    cpdef uint32_t get(self, uint32_t) noexcept
    cdef void set(self, uint32_t, uint32_t) noexcept
    cdef uint32_t get_sp(self) noexcept
    cdef void set_sp(self, uint32_t) noexcept
    cdef uint32_t get_lr(self) noexcept
    cdef void set_lr(self, uint32_t) noexcept
    cdef uint32_t get_pc(self) noexcept
    cdef void set_pc(self, uint32_t) noexcept
    cdef void switch_mode(self, int) noexcept
    cdef int get_bank_index(self, int) noexcept
    cdef ProgramStatusRegister get_banked_spsr(self, int) noexcept


cdef class ProgramStatusRegister:
    cdef uint32_t reg
    cdef int get_mode(self) noexcept
    cdef void set_mode(self, int) noexcept
    cdef int get_state(self) noexcept
    cdef void set_state(self, int) noexcept
    cdef bint get_sign_flag(self) noexcept
    cdef void set_sign_flag(self, bint) noexcept
    cdef bint get_zero_flag(self) noexcept
    cdef void set_zero_flag(self, bint) noexcept
    cdef bint get_carry_flag(self) noexcept
    cdef void set_carry_flag(self, bint) noexcept
    cdef bint get_overflow_flag(self) noexcept
    cdef void set_overflow_flag(self, bint) noexcept
    cdef bint get_sticky_overflow_flag(self) noexcept
    cdef void set_sticky_overflow_flag(self, bint) noexcept
    cdef bint get_irq_disable(self) noexcept
    cdef void set_irq_disable(self, bint) noexcept
    cdef bint get_fiq_disable(self) noexcept
    cdef void set_fiq_disable(self, bint) noexcept



cimport cython

from libc.stdint cimport uint8_t, uint32_t

from pyboy_advance.interrupt_controller cimport InterruptController
from pyboy_advance.keypad cimport Keypad
from pyboy_advance.memory.constants cimport IOAddress
from pyboy_advance.memory.dma cimport DMAController
from pyboy_advance.memory.memory cimport Memory
from pyboy_advance.ppu.constants cimport WindowIndex
from pyboy_advance.ppu.ppu cimport PPU
from pyboy_advance.timer cimport Timers
from pyboy_advance.utils cimport get_bit

cdef class IO:
    cdef Memory memory
    cdef InterruptController interrupt_controller
    cdef DMAController dma_controller
    cdef Timers timers
    cdef PPU ppu
    cdef Keypad keypad
    cdef uint32_t reg_soundbias

    cdef uint32_t read_32(self, uint32_t) noexcept
    cdef uint32_t read_16(self, uint32_t) noexcept
    cdef uint32_t read_8(self, uint32_t) noexcept

    cdef void write_32(self, uint32_t, uint32_t) noexcept

    @cython.locals(mask=uint32_t)
    cdef void write_16(self, uint32_t, uint32_t) noexcept

    cdef void write_8(self, uint32_t, uint32_t) noexcept



import ctypes
try:
    import numpy as np
except ImportError:
    np = None
try:
    import PIL.Image as PILImage
except ImportError:
    PILImage = None

class Screen:
    """
    Provides access to the frame buffer of PyBoy Advance.
    """

    def __init__(self, ppu: PPU):
        self.ppu = ppu

    @property
    def ndarray(self) -> 'np.ndarray':
        """
        Get the current frame buffer in a NumPy ``ndarray``.

        :return: RGB image in ``ndarray`` of bytes with shape (160, 240, 3)
        """
        buffer_type = ctypes.POINTER(ctypes.c_uint16 * DISPLAY_HEIGHT * DISPLAY_WIDTH)
        buffer = ctypes.cast(self.ppu.get_frame_buffer_ptr(), buffer_type).contents
        img_bgr555 = np.frombuffer(buffer, dtype=np.uint16).reshape((DISPLAY_HEIGHT, DISPLAY_WIDTH))
        b = img_bgr555 >> 10 & 31
        g = img_bgr555 >> 5 & 31
        r = img_bgr555 & 31
        r = (r * 255 // 31).astype(np.uint8)
        g = (g * 255 // 31).astype(np.uint8)
        b = (b * 255 // 31).astype(np.uint8)
        return np.stack([r, g, b], axis=-1)

    @property
    def image(self) -> 'PILImage.Image':
        """
        Get a PIL Image of the current frame buffer.

        :return: RGB image of (240, 160) pixels
        """
        return PILImage.fromarray(self.ndarray, 'RGB')


import pytest

from pyboy_advance.constants import Key


def select_test_num(emulator, test_num):
    for _ in range(test_num):
        emulator.press_key(Key.DPAD_DOWN)
        emulator.frame(5)
        emulator.release_key(Key.DPAD_DOWN)
        emulator.frame(5)

    emulator.press_key(Key.BUTTON_A)
    emulator.frame(5)
    emulator.release_key(Key.BUTTON_A)
    emulator.frame(5)


@pytest.mark.skip
@pytest.mark.rom_url("https://s3.amazonaws.com/mgba/suite-latest.zip")
def test_mgba_memory(emulator, assert_expected_image):
    emulator.frame(20)
    select_test_num(emulator, 0)
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.skip
@pytest.mark.rom_url("https://s3.amazonaws.com/mgba/suite-latest.zip")
def test_mgba_io(emulator, assert_expected_image):
    emulator.frame(20)
    select_test_num(emulator, 1)
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.skip
@pytest.mark.rom_url("https://s3.amazonaws.com/mgba/suite-latest.zip")
def test_mgba_timing(emulator, assert_expected_image):
    emulator.frame(20)
    select_test_num(emulator, 2)
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.skip
@pytest.mark.rom_url("https://s3.amazonaws.com/mgba/suite-latest.zip")
def test_mgba_timer_count_up(emulator, assert_expected_image):
    emulator.frame(20)
    select_test_num(emulator, 3)
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.skip
@pytest.mark.rom_url("https://s3.amazonaws.com/mgba/suite-latest.zip")
def test_mgba_timer_irq(emulator, assert_expected_image):
    emulator.frame(20)
    select_test_num(emulator, 4)
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.rom_url("https://s3.amazonaws.com/mgba/suite-latest.zip")
def test_mgba_shifter(emulator, assert_expected_image):
    emulator.frame(20)
    select_test_num(emulator, 5)
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.rom_url("https://s3.amazonaws.com/mgba/suite-latest.zip")
def test_mgba_carry(emulator, assert_expected_image):
    emulator.frame(20)
    select_test_num(emulator, 6)
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.skip
@pytest.mark.rom_url("https://s3.amazonaws.com/mgba/suite-latest.zip")
def test_mgba_multiply_long(emulator, assert_expected_image):
    emulator.frame(20)
    select_test_num(emulator, 7)
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.skip
@pytest.mark.rom_url("https://s3.amazonaws.com/mgba/suite-latest.zip")
def test_mgba_bios(emulator, assert_expected_image):
    emulator.frame(20)
    select_test_num(emulator, 8)
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)


@pytest.mark.skip
@pytest.mark.rom_url("https://s3.amazonaws.com/mgba/suite-latest.zip")
def test_mgba_dma(emulator, assert_expected_image):
    emulator.frame(20)
    select_test_num(emulator, 9)
    emulator.frame(20)
    assert_expected_image(emulator.screen.image)



def get_bit(num: int, i: int) -> bint:
    return num >> i & 1

def get_bits(num: int, start: int, end: int) -> int:
    num_bits = end - start + 1
    mask = (1 << num_bits) - 1
    return num >> start & mask

def set_bit(num: int, i: int, bit: bint) -> int:
    return num & ~(1 << i) | bit << i

def sign_32(num: int) -> bint:
    return get_bit(num, 31)

def sign_24(num: int) -> bint:
    return get_bit(num, 23)

def sign_23(num: int) -> bint:
    return get_bit(num, 22)

def sign_16(num: int) -> bint:
    return get_bit(num, 15)

def sign_12(num: int) -> bint:
    return get_bit(num, 11)

def sign_9(num: int) -> bint:
    return get_bit(num, 8)

def sign_8(num: int) -> bint:
    return get_bit(num, 7)

def extend_sign_32(num: int) -> int:
    extension = 18446744069414584320
    return num | extension if sign_32(num) else num

def extend_sign_24(num: int) -> int:
    extension = 4278190080
    return num | extension if sign_24(num) else num

def extend_sign_23(num: int) -> int:
    extension = 4286578688
    return num | extension if sign_23(num) else num

def extend_sign_16(num: int) -> int:
    extension = 4294901760
    return num | extension if sign_16(num) else num

def extend_sign_12(num: int) -> int:
    extension = 4294963200
    return num | extension if sign_12(num) else num

def extend_sign_9(num: int) -> int:
    extension = 4294966784
    return num | extension if sign_9(num) else num

def extend_sign_8(num: int) -> int:
    extension = 4294967040
    return num | extension if sign_8(num) else num

def add_32(op1: int, op2: int) -> int:
    mask = 4294967295
    return op1 + op2 & mask

def ror_32(num: int, amount: int) -> int:
    """
    Rotate right (ROR)

    :param num: 32-bit unsigned integer to rotate
    :param amount: number of bits to rotate to the right (must be <= 32)
    :return: num rotated right by amount bits
    """
    mask = 4294967295
    return (num >> amount | num << 32 - amount) & mask

def array_read_32(arr: array, address: int) -> int:
    """Read a little-endian 32-bit value from an array of bytes"""
    b0 = arr[address]
    b1 = arr[address + 1]
    b2 = arr[address + 2]
    b3 = arr[address + 3]
    return b0 | b1 << 8 | b2 << 16 | b3 << 24

def array_read_16(arr: array, address: int) -> int:
    """Read a little-endian 16-bit value from an array of bytes"""
    b0 = arr[address]
    b1 = arr[address + 1]
    return b0 | b1 << 8

def array_write_32(arr: array, address: int, value: int):
    """Write a 32-bit value to the given array of bytes in little-endian format"""
    arr[address] = value & 255
    arr[address + 1] = value >> 8 & 255
    arr[address + 2] = value >> 16 & 255
    arr[address + 3] = value >> 24 & 255

def array_write_16(arr: array, address: int, value: int):
    """Write a 16-bit value to the given array of bytes in little-endian format"""
    arr[address] = value & 255
    arr[address + 1] = value >> 8 & 255


import os
import re


def preprocess_cython_guards(file_paths: list[str | os.PathLike]) -> list[str | os.PathLike]:
    """
    Remove code blocks guarded by '# ifndef CYTHON' ... '# endif'.
    """
    start_re = re.compile(r"^\s*#\s*ifndef\s+CYTHON\b", flags=re.IGNORECASE)
    end_re = re.compile(r"^\s*#\s*endif\b", flags=re.IGNORECASE)

    for file_path in file_paths:
        with open(file_path, "r", encoding="utf-8") as f:
            lines = f.readlines()

        out_lines = []
        in_guard = False

        for line in lines:
            if not in_guard and start_re.match(line):
                in_guard = True
                continue
            if in_guard and end_re.match(line):
                in_guard = False
                continue
            if not in_guard:
                out_lines.append(line)

        with open(file_path, "w", encoding="utf-8") as f:
            f.write("".join(out_lines))

    return file_paths



class Timers:

    def __init__(self, scheduler: Scheduler, interrupt_controller: InterruptController):
        self.timer_3 = Timer(scheduler, interrupt_controller, Interrupt.TIMER_3, None)
        self.timer_2 = Timer(scheduler, interrupt_controller, Interrupt.TIMER_2, self.timer_3)
        self.timer_1 = Timer(scheduler, interrupt_controller, Interrupt.TIMER_1, self.timer_2)
        self.timer_0 = Timer(scheduler, interrupt_controller, Interrupt.TIMER_0, self.timer_1)

class Timer:

    def __init__(self, scheduler: Scheduler, interrupt_controller: InterruptController, interrupt: Interrupt, next_timer: Timer | None):
        self.WRITE_TIMER_REGISTER_DELAY = 1
        self.START_TIMER_DELAY = 1
        self.FREQUENCIES = [1, 64, 256, 1024]
        self.scheduler = scheduler
        self.interrupt_controller = interrupt_controller
        self.interrupt = interrupt
        self.next_timer = next_timer
        self._counter = 0
        self._reload_value = 0
        self._control_reg = 0
        self._pending_reload_value = 0
        self._pending_control_reg = 0
        self._overflow_event: Event | None = None

    def get_counter(self) -> int:
        if self.get_timer_enabled() and (not self.get_count_up()):
            self._update_counter()
        return self._counter

    def set_counter(self, value: int):
        self._pending_reload_value = value
        self.scheduler.schedule(self._write_reload_value, self.WRITE_TIMER_REGISTER_DELAY, EventTrigger.IMMEDIATELY)

    def get_control_reg(self) -> int:
        return self._control_reg

    def set_control_reg(self, value: int):
        self._pending_control_reg = value
        self.scheduler.schedule(self._write_control_reg, self.WRITE_TIMER_REGISTER_DELAY, EventTrigger.IMMEDIATELY)

    def get_frequency(self) -> int:
        return self.FREQUENCIES[get_bits(self._control_reg, 0, 1)]

    def get_count_up(self) -> bint:
        return get_bit(self._control_reg, 2)

    def set_count_up(self, value: bint):
        set_bit(self._control_reg, 2, value)

    def get_irq_enabled(self) -> bint:
        return get_bit(self._control_reg, 6)

    def get_timer_enabled(self) -> bint:
        return get_bit(self._control_reg, 7)

    def _write_reload_value(self):
        self._reload_value = self._pending_reload_value

    def _write_control_reg(self):
        old_count_up = self.get_count_up()
        old_timer_enabled = self.get_timer_enabled()
        self._control_reg = self._pending_control_reg
        new_count_up = self.get_count_up()
        new_timer_enabled = self.get_timer_enabled()
        if new_timer_enabled and (not old_timer_enabled):
            self._start()
        elif not new_timer_enabled and old_timer_enabled:
            self._stop()
        elif new_count_up and (not old_count_up):
            self._stop()
        elif not new_count_up and old_count_up:
            self._start()

    def _start(self):
        if not self.get_count_up():
            self._counter = self._reload_value
            self._schedule_overflow_event(self.START_TIMER_DELAY)
        elif self._overflow_event:
            self._schedule_overflow_event(0)

    def _stop(self):
        self._update_counter()
        if self._overflow_event:
            self._overflow_event.cancelled = True
            self._overflow_event = None

    def _update_counter(self):
        if self._overflow_event:
            elapsed = self.scheduler.cycles - self._overflow_event.time & 65535
            self._counter = elapsed // self.get_frequency()

    def _schedule_overflow_event(self, start_delay: int):
        delay = (65536 - self._counter) * self.get_frequency() + start_delay
        if self._overflow_event:
            self._overflow_event.cancelled = True
            self._overflow_event = None
        self._overflow_event = self.scheduler.schedule(self._overflow, delay, EventTrigger.IMMEDIATELY)

    def _overflow(self):
        self._counter = self._reload_value
        if not self.get_count_up():
            self._schedule_overflow_event(0)
        elif self._overflow_event:
            self._overflow_event.cancelled = True
            self._overflow_event = None
        if self.get_irq_enabled():
            self.interrupt_controller.signal(self.interrupt)
        if self.next_timer and self.next_timer.get_count_up() and self.next_timer.get_timer_enabled():
            self.next_timer._counter += 1
            if self.next_timer._counter == 65536:
                self.next_timer._overflow()


from ctypes import c_void_p
import sdl2.ext
DEFAULT_KEY_DOWN_MAP = {sdl2.SDLK_a: WindowEvent.PRESS_BUTTON_A, sdl2.SDLK_s: WindowEvent.PRESS_BUTTON_B, sdl2.SDLK_BACKSPACE: WindowEvent.PRESS_BUTTON_SELECT, sdl2.SDLK_RETURN: WindowEvent.PRESS_BUTTON_START, sdl2.SDLK_RIGHT: WindowEvent.PRESS_DPAD_RIGHT, sdl2.SDLK_LEFT: WindowEvent.PRESS_DPAD_LEFT, sdl2.SDLK_UP: WindowEvent.PRESS_DPAD_UP, sdl2.SDLK_DOWN: WindowEvent.PRESS_DPAD_DOWN, sdl2.SDLK_w: WindowEvent.PRESS_SHOULDER_RIGHT, sdl2.SDLK_q: WindowEvent.PRESS_SHOULDER_LEFT}
DEFAULT_KEY_UP_MAP = {sdl2.SDLK_a: WindowEvent.RELEASE_BUTTON_A, sdl2.SDLK_s: WindowEvent.RELEASE_BUTTON_B, sdl2.SDLK_BACKSPACE: WindowEvent.RELEASE_BUTTON_SELECT, sdl2.SDLK_RETURN: WindowEvent.RELEASE_BUTTON_START, sdl2.SDLK_RIGHT: WindowEvent.RELEASE_DPAD_RIGHT, sdl2.SDLK_LEFT: WindowEvent.RELEASE_DPAD_LEFT, sdl2.SDLK_UP: WindowEvent.RELEASE_DPAD_UP, sdl2.SDLK_DOWN: WindowEvent.RELEASE_DPAD_DOWN, sdl2.SDLK_w: WindowEvent.RELEASE_SHOULDER_RIGHT, sdl2.SDLK_q: WindowEvent.RELEASE_SHOULDER_LEFT, sdl2.SDLK_ESCAPE: WindowEvent.QUIT, sdl2.SDLK_F11: WindowEvent.FULLSCREEN}

class Window:

    def __init__(self):
        self._events = []
        self._key_down_map = DEFAULT_KEY_DOWN_MAP
        self._key_up_map = DEFAULT_KEY_UP_MAP
        self._window_scale = 3

    def __enter__(self):
        if sdl2.SDL_InitSubSystem(sdl2.SDL_INIT_VIDEO | sdl2.SDL_INIT_GAMECONTROLLER) < 0:
            raise ValueError(f'SDL_InitSubSystem failed: {sdl2.SDL_GetError().decode()}')
        self._sdl_window = sdl2.SDL_CreateWindow(b'PyBoy Advance', sdl2.SDL_WINDOWPOS_CENTERED, sdl2.SDL_WINDOWPOS_CENTERED, DISPLAY_WIDTH * self._window_scale, DISPLAY_HEIGHT * self._window_scale, sdl2.SDL_WINDOW_RESIZABLE)
        self._sdl_renderer = sdl2.SDL_CreateRenderer(self._sdl_window, -1, sdl2.SDL_RENDERER_ACCELERATED)
        sdl2.SDL_RenderSetLogicalSize(self._sdl_renderer, DISPLAY_WIDTH, DISPLAY_HEIGHT)
        self._sdl_texture_buffer = sdl2.SDL_CreateTexture(self._sdl_renderer, sdl2.SDL_PIXELFORMAT_BGR555, sdl2.SDL_TEXTUREACCESS_STREAMING, DISPLAY_WIDTH, DISPLAY_HEIGHT)
        sdl2.SDL_ShowWindow(self._sdl_window)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sdl2.SDL_DestroyWindow(self._sdl_window)
        sdl2.SDL_QuitSubSystem(sdl2.SDL_INIT_VIDEO | sdl2.SDL_INIT_GAMECONTROLLER)

    def get_fullscreen(self) -> bool:
        flags = sdl2.SDL_GetWindowFlags(self._sdl_window)
        return flags & sdl2.SDL_WINDOW_FULLSCREEN_DESKTOP

    def set_fullscreen(self, value: bool):
        sdl2.SDL_SetWindowFullscreen(self._sdl_window, sdl2.SDL_WINDOW_FULLSCREEN_DESKTOP if value else 0)

    def get_events(self) -> list[WindowEvent]:
        self._events.clear()
        for event in sdl2.ext.get_events():
            if event.type == sdl2.SDL_QUIT:
                self._events.append(WindowEvent.QUIT)
            elif event.type == sdl2.SDL_KEYDOWN:
                self._events.append(self._key_down_map.get(event.key.keysym.sym, WindowEvent.WindowEvent_NONE))
            elif event.type == sdl2.SDL_KEYUP:
                self._events.append(self._key_up_map.get(event.key.keysym.sym, WindowEvent.WindowEvent_NONE))
        return self._events

    def render(self, frame_buffer_ptr: c_void_p):
        sdl2.SDL_UpdateTexture(self._sdl_texture_buffer, None, frame_buffer_ptr, DISPLAY_WIDTH * COLOUR_SIZE)
        sdl2.SDL_RenderClear(self._sdl_renderer)
        sdl2.SDL_RenderCopy(self._sdl_renderer, self._sdl_texture_buffer, None, None)
        sdl2.SDL_RenderPresent(self._sdl_renderer)


def arm_alu(cpu: CPU, instr: int):
    """Execute a Data Processing (ALU) instruction (AND, EOR, ADD, SUB, MOV, etc.)"""
    rn = get_bits(instr, 16, 19)
    rd = get_bits(instr, 12, 15)
    shift_carry = cpu.regs.cpsr.get_carry_flag()
    set_cond_codes = get_bit(instr, 20)
    immediate = get_bit(instr, 25)
    early_advance_pc = False
    if immediate:
        op1 = cpu.regs.get(rn)
        op2 = get_bits(instr, 0, 7)
        ror_amount = get_bits(instr, 8, 11) * 2
        if ror_amount > 0:
            shift_carry = get_bit(op2, ror_amount - 1)
            op2 = ror_32(op2, ror_amount)
    else:
        rm = get_bits(instr, 0, 3)
        shift = get_bits(instr, 4, 11)
        if get_bit(shift, 0):
            cpu.advance_pc_arm()
            early_advance_pc = True
            cpu.scheduler.idle(1)
        else:
            pass
        op1 = cpu.regs.get(rn)
        op2, shift_carry = cpu.decode_and_compute_shift(cpu.regs.get(rm), shift)
    opcode = get_bits(instr, 21, 24)
    if opcode == ALUOpcode.AND:
        arm_alu_and(cpu, op1, op2, rd, set_cond_codes, shift_carry)
    elif opcode == ALUOpcode.EOR:
        arm_alu_eor(cpu, op1, op2, rd, set_cond_codes, shift_carry)
    elif opcode == ALUOpcode.SUB:
        arm_alu_sub(cpu, op1, op2, rd, set_cond_codes)
    elif opcode == ALUOpcode.RSB:
        arm_alu_rsb(cpu, op1, op2, rd, set_cond_codes)
    elif opcode == ALUOpcode.ADD:
        arm_alu_add(cpu, op1, op2, rd, set_cond_codes)
    elif opcode == ALUOpcode.ADC:
        arm_alu_adc(cpu, op1, op2, rd, set_cond_codes)
    elif opcode == ALUOpcode.SBC:
        arm_alu_sbc(cpu, op1, op2, rd, set_cond_codes)
    elif opcode == ALUOpcode.RSC:
        arm_alu_rsc(cpu, op1, op2, rd, set_cond_codes)
    elif opcode == ALUOpcode.TST:
        arm_alu_tst(cpu, op1, op2, shift_carry)
    elif opcode == ALUOpcode.TEQ:
        arm_alu_teq(cpu, op1, op2, shift_carry)
    elif opcode == ALUOpcode.CMP:
        arm_alu_cmp(cpu, op1, op2)
    elif opcode == ALUOpcode.CMN:
        arm_alu_cmn(cpu, op1, op2)
    elif opcode == ALUOpcode.ORR:
        arm_alu_orr(cpu, op1, op2, rd, set_cond_codes, shift_carry)
    elif opcode == ALUOpcode.MOV:
        arm_alu_mov(cpu, op2, rd, set_cond_codes, shift_carry)
    elif opcode == ALUOpcode.BIC:
        arm_alu_bic(cpu, op1, op2, rd, set_cond_codes, shift_carry)
    elif opcode == ALUOpcode.MVN:
        arm_alu_mvn(cpu, op2, rd, set_cond_codes, shift_carry)
    if rd == cpu.regs.PC:
        if set_cond_codes:
            new_cpsr_reg = cpu.regs.spsr.reg
            new_cpsr_mode = cpu.regs.spsr.get_mode()
            cpu.switch_mode(new_cpsr_mode)
            cpu.regs.cpsr.reg = new_cpsr_reg
        if opcode not in [ALUOpcode.TST, ALUOpcode.TEQ, ALUOpcode.CMP, ALUOpcode.CMN]:
            cpu.flush_pipeline()
    elif not early_advance_pc:
        cpu.advance_pc_arm()

def arm_alu_and(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint):
    cpu.regs.set(rd, arm_alu_and_impl(cpu, op1, op2, rd, set_cond_codes, shift_carry))

def arm_alu_and_impl(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint) -> int:
    result = op1 & op2
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.set_sign_flag(sign_32(result))
        cpu.regs.cpsr.set_zero_flag(result == 0)
        cpu.regs.cpsr.set_carry_flag(shift_carry)
    return result

def arm_alu_eor(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint):
    cpu.regs.set(rd, arm_alu_eor_impl(cpu, op1, op2, rd, set_cond_codes, shift_carry))

def arm_alu_eor_impl(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint) -> int:
    result = op1 ^ op2
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.set_sign_flag(sign_32(result))
        cpu.regs.cpsr.set_zero_flag(result == 0)
        cpu.regs.cpsr.set_carry_flag(shift_carry)
    return result

def arm_alu_sub(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint):
    cpu.regs.set(rd, arm_alu_sub_impl(cpu, op1, op2, rd, set_cond_codes))

def arm_alu_sub_impl(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint) -> int:
    mask = 4294967295
    result = op1 - op2 & mask
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.set_sign_flag(sign_32(result))
        cpu.regs.cpsr.set_zero_flag(result == 0)
        cpu.regs.cpsr.set_carry_flag(op1 >= op2)
        cpu.regs.cpsr.set_overflow_flag(sign_32((op1 ^ op2) & (op1 ^ result)))
    return result

def arm_alu_rsb(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint):
    arm_alu_sub(cpu, op2, op1, rd, set_cond_codes)

def arm_alu_add(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint):
    cpu.regs.set(rd, arm_alu_add_impl(cpu, op1, op2, rd, set_cond_codes))

def arm_alu_add_impl(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint) -> int:
    mask = 4294967295
    result = op1 + op2 & mask
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.set_sign_flag(sign_32(result))
        cpu.regs.cpsr.set_zero_flag(result == 0)
        cpu.regs.cpsr.set_carry_flag(result < op1)
        cpu.regs.cpsr.set_overflow_flag(sign_32(~(op1 ^ op2) & (op2 ^ result)))
    return result

def arm_alu_adc(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint):
    mask = 4294967295
    temp = op1 + op2 & mask
    carry_1 = temp < op1
    result = temp + cpu.regs.cpsr.get_carry_flag() & mask
    carry_2 = result < temp
    cpu.regs.set(rd, result)
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.set_sign_flag(sign_32(result))
        cpu.regs.cpsr.set_zero_flag(result == 0)
        cpu.regs.cpsr.set_carry_flag(carry_1 or carry_2)
        cpu.regs.cpsr.set_overflow_flag(sign_32(~(op1 ^ op2) & (op2 ^ result)))

def arm_alu_sbc(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint):
    mask = 4294967295
    borrow_in = 1 - cpu.regs.cpsr.get_carry_flag()
    temp = op1 - op2 & mask
    borrow_1 = op1 < op2
    result = temp - borrow_in & mask
    borrow_2 = temp < borrow_in
    cpu.regs.set(rd, result)
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.set_sign_flag(sign_32(result))
        cpu.regs.cpsr.set_zero_flag(result == 0)
        cpu.regs.cpsr.set_carry_flag(not (borrow_1 or borrow_2))
        cpu.regs.cpsr.set_overflow_flag(sign_32((op1 ^ op2) & (op1 ^ result)))

def arm_alu_rsc(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint):
    arm_alu_sbc(cpu, op2, op1, rd, set_cond_codes)

def arm_alu_tst(cpu: CPU, op1: int, op2: int, shift_carry: bint):
    arm_alu_and_impl(cpu, op1, op2, 0, True, shift_carry)

def arm_alu_teq(cpu: CPU, op1: int, op2: int, shift_carry: bint):
    arm_alu_eor_impl(cpu, op1, op2, 0, True, shift_carry)

def arm_alu_cmp(cpu: CPU, op1: int, op2: int):
    arm_alu_sub_impl(cpu, op1, op2, 0, True)

def arm_alu_cmn(cpu: CPU, op1: int, op2: int):
    arm_alu_add_impl(cpu, op1, op2, 0, True)

def arm_alu_orr(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint):
    result = op1 | op2
    cpu.regs.set(rd, result)
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.set_sign_flag(sign_32(result))
        cpu.regs.cpsr.set_zero_flag(result == 0)
        cpu.regs.cpsr.set_carry_flag(shift_carry)

def arm_alu_mov(cpu: CPU, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint):
    cpu.regs.set(rd, op2)
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.set_sign_flag(sign_32(op2))
        cpu.regs.cpsr.set_zero_flag(op2 == 0)
        cpu.regs.cpsr.set_carry_flag(shift_carry)

def arm_alu_bic(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint):
    result = op1 & ~op2
    cpu.regs.set(rd, result)
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.set_sign_flag(sign_32(result))
        cpu.regs.cpsr.set_zero_flag(result == 0)
        cpu.regs.cpsr.set_carry_flag(shift_carry)

def arm_alu_mvn(cpu: CPU, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint):
    mask = 4294967295
    result = ~op2 & mask
    cpu.regs.set(rd, result)
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.set_sign_flag(sign_32(result))
        cpu.regs.cpsr.set_zero_flag(result == 0)
        cpu.regs.cpsr.set_carry_flag(shift_carry)


# ifndef CYTHON
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.cpu.cpu import CPU

from pyboy_advance.cpu.arm.constants import ALUOpcode
from pyboy_advance.utils import get_bits, get_bit, ror_32, sign_32, bint
# endif


def arm_alu(cpu: CPU, instr: int):
    """Execute a Data Processing (ALU) instruction (AND, EOR, ADD, SUB, MOV, etc.)"""

    rn = get_bits(instr, 16, 19)
    rd = get_bits(instr, 12, 15)

    shift_carry = cpu.regs.cpsr.carry_flag

    set_cond_codes = get_bit(instr, 20)
    immediate = get_bit(instr, 25)

    early_advance_pc = False

    if immediate:  # Immediate value as 2nd operand
        op1 = cpu.regs.get(rn)
        op2 = get_bits(instr, 0, 7)
        ror_amount = get_bits(instr, 8, 11) * 2

        if ror_amount > 0:
            shift_carry = get_bit(op2, ror_amount - 1)
            op2 = ror_32(op2, ror_amount)

    else:  # Register value as 2nd operand
        rm = get_bits(instr, 0, 3)
        shift = get_bits(instr, 4, 11)

        # When using PC as an operand, the returned value depends on the instruction.
        # If shifting by register (shift[0] = 1) and using a register value as op2,
        # then the operand should be PC + 12.
        if get_bit(shift, 0):
            # Advance PC by 4 so that cpu.regs.pc returns PC + 12
            cpu.advance_pc_arm()
            early_advance_pc = True
            cpu.scheduler.idle(1)
        else:
            # Otherwise, the operand should be PC + 8
            # (which is what cpu.regs.pc returns normally).
            pass

        op1 = cpu.regs.get(rn)
        op2, shift_carry = cpu.decode_and_compute_shift(cpu.regs.get(rm), shift)

    opcode = get_bits(instr, 21, 24)
    if opcode == ALUOpcode.AND:
        arm_alu_and(cpu, op1, op2, rd, set_cond_codes, shift_carry)
    elif opcode == ALUOpcode.EOR:
        arm_alu_eor(cpu, op1, op2, rd, set_cond_codes, shift_carry)
    elif opcode == ALUOpcode.SUB:
        arm_alu_sub(cpu, op1, op2, rd, set_cond_codes)
    elif opcode == ALUOpcode.RSB:
        arm_alu_rsb(cpu, op1, op2, rd, set_cond_codes)
    elif opcode == ALUOpcode.ADD:
        arm_alu_add(cpu, op1, op2, rd, set_cond_codes)
    elif opcode == ALUOpcode.ADC:
        arm_alu_adc(cpu, op1, op2, rd, set_cond_codes)
    elif opcode == ALUOpcode.SBC:
        arm_alu_sbc(cpu, op1, op2, rd, set_cond_codes)
    elif opcode == ALUOpcode.RSC:
        arm_alu_rsc(cpu, op1, op2, rd, set_cond_codes)
    elif opcode == ALUOpcode.TST:
        arm_alu_tst(cpu, op1, op2, shift_carry)
    elif opcode == ALUOpcode.TEQ:
        arm_alu_teq(cpu, op1, op2, shift_carry)
    elif opcode == ALUOpcode.CMP:
        arm_alu_cmp(cpu, op1, op2)
    elif opcode == ALUOpcode.CMN:
        arm_alu_cmn(cpu, op1, op2)
    elif opcode == ALUOpcode.ORR:
        arm_alu_orr(cpu, op1, op2, rd, set_cond_codes, shift_carry)
    elif opcode == ALUOpcode.MOV:
        arm_alu_mov(cpu, op2, rd, set_cond_codes, shift_carry)
    elif opcode == ALUOpcode.BIC:
        arm_alu_bic(cpu, op1, op2, rd, set_cond_codes, shift_carry)
    elif opcode == ALUOpcode.MVN:
        arm_alu_mvn(cpu, op2, rd, set_cond_codes, shift_carry)

    # When S bit = 1 (set_cond_codes) and Rd = PC, the result of operation is stored in PC
    # and SPSR of the current mode is moved to CPSR
    if rd == cpu.regs.PC:
        if set_cond_codes:
            new_cpsr_reg = cpu.regs.spsr.reg
            new_cpsr_mode = cpu.regs.spsr.mode
            cpu.switch_mode(new_cpsr_mode)
            cpu.regs.cpsr.reg = new_cpsr_reg

        # Read only operations do not flush the pipeline
        if opcode not in [ALUOpcode.TST, ALUOpcode.TEQ, ALUOpcode.CMP, ALUOpcode.CMN]:
            cpu.flush_pipeline()

    elif not early_advance_pc:
        cpu.advance_pc_arm()


def arm_alu_and(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint):
    cpu.regs.set(rd, arm_alu_and_impl(cpu, op1, op2, rd, set_cond_codes, shift_carry))


def arm_alu_and_impl(
    cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint
) -> int:
    result = op1 & op2
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.sign_flag = sign_32(result)
        cpu.regs.cpsr.zero_flag = result == 0
        cpu.regs.cpsr.carry_flag = shift_carry
    return result


def arm_alu_eor(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint):
    cpu.regs.set(rd, arm_alu_eor_impl(cpu, op1, op2, rd, set_cond_codes, shift_carry))


def arm_alu_eor_impl(
    cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint
) -> int:
    result = op1 ^ op2
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.sign_flag = sign_32(result)
        cpu.regs.cpsr.zero_flag = result == 0
        cpu.regs.cpsr.carry_flag = shift_carry
    return result


def arm_alu_sub(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint):
    cpu.regs.set(rd, arm_alu_sub_impl(cpu, op1, op2, rd, set_cond_codes))


def arm_alu_sub_impl(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint) -> int:
    mask = 0xFFFFFFFF
    result = (op1 - op2) & mask
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.sign_flag = sign_32(result)
        cpu.regs.cpsr.zero_flag = result == 0
        cpu.regs.cpsr.carry_flag = op1 >= op2  # Carry = no borrow
        cpu.regs.cpsr.overflow_flag = sign_32((op1 ^ op2) & (op1 ^ result))
    return result


def arm_alu_rsb(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint):
    arm_alu_sub(cpu, op2, op1, rd, set_cond_codes)


def arm_alu_add(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint):
    cpu.regs.set(rd, arm_alu_add_impl(cpu, op1, op2, rd, set_cond_codes))


def arm_alu_add_impl(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint) -> int:
    mask = 0xFFFFFFFF
    result = (op1 + op2) & mask
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.sign_flag = sign_32(result)
        cpu.regs.cpsr.zero_flag = result == 0
        cpu.regs.cpsr.carry_flag = result < op1
        cpu.regs.cpsr.overflow_flag = sign_32(~(op1 ^ op2) & (op2 ^ result))
    return result


def arm_alu_adc(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint):
    mask = 0xFFFFFFFF
    temp = (op1 + op2) & mask
    carry_1 = temp < op1
    result = (temp + cpu.regs.cpsr.carry_flag) & mask
    carry_2 = result < temp
    cpu.regs.set(rd, result)
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.sign_flag = sign_32(result)
        cpu.regs.cpsr.zero_flag = result == 0
        cpu.regs.cpsr.carry_flag = carry_1 or carry_2
        cpu.regs.cpsr.overflow_flag = sign_32(~(op1 ^ op2) & (op2 ^ result))


def arm_alu_sbc(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint):
    mask = 0xFFFFFFFF
    borrow_in = 1 - cpu.regs.cpsr.carry_flag  # Carry = no borrow, so subtract 0
    temp = (op1 - op2) & mask
    borrow_1 = op1 < op2
    result = (temp - borrow_in) & mask
    borrow_2 = temp < borrow_in
    cpu.regs.set(rd, result)
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.sign_flag = sign_32(result)
        cpu.regs.cpsr.zero_flag = result == 0
        cpu.regs.cpsr.carry_flag = not (borrow_1 or borrow_2)
        cpu.regs.cpsr.overflow_flag = sign_32((op1 ^ op2) & (op1 ^ result))


def arm_alu_rsc(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint):
    arm_alu_sbc(cpu, op2, op1, rd, set_cond_codes)


def arm_alu_tst(cpu: CPU, op1: int, op2: int, shift_carry: bint):
    arm_alu_and_impl(cpu, op1, op2, 0, True, shift_carry)


def arm_alu_teq(cpu: CPU, op1: int, op2: int, shift_carry: bint):
    arm_alu_eor_impl(cpu, op1, op2, 0, True, shift_carry)


def arm_alu_cmp(cpu: CPU, op1: int, op2: int):
    arm_alu_sub_impl(cpu, op1, op2, 0, True)


def arm_alu_cmn(cpu: CPU, op1: int, op2: int):
    arm_alu_add_impl(cpu, op1, op2, 0, True)


def arm_alu_orr(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint):
    result = op1 | op2
    cpu.regs.set(rd, result)
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.sign_flag = sign_32(result)
        cpu.regs.cpsr.zero_flag = result == 0
        cpu.regs.cpsr.carry_flag = shift_carry


def arm_alu_mov(cpu: CPU, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint):
    cpu.regs.set(rd, op2)
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.sign_flag = sign_32(op2)
        cpu.regs.cpsr.zero_flag = op2 == 0
        cpu.regs.cpsr.carry_flag = shift_carry


def arm_alu_bic(cpu: CPU, op1: int, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint):
    result = op1 & ~op2
    cpu.regs.set(rd, result)
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.sign_flag = sign_32(result)
        cpu.regs.cpsr.zero_flag = result == 0
        cpu.regs.cpsr.carry_flag = shift_carry


def arm_alu_mvn(cpu: CPU, op2: int, rd: int, set_cond_codes: bint, shift_carry: bint):
    mask = 0xFFFFFFFF
    result = ~op2 & mask
    cpu.regs.set(rd, result)
    if set_cond_codes and rd != cpu.regs.PC:
        cpu.regs.cpsr.sign_flag = sign_32(result)
        cpu.regs.cpsr.zero_flag = result == 0
        cpu.regs.cpsr.carry_flag = shift_carry



from libc.stdint cimport uint32_t

from pyboy_advance.cpu.cpu cimport CPU, InstrHandler, InstrPattern
from pyboy_advance.cpu.arm.alu cimport arm_alu
from pyboy_advance.cpu.arm.bdt cimport arm_block_data_transfer
from pyboy_advance.cpu.arm.branch cimport arm_branch_exchange, arm_branch
from pyboy_advance.cpu.arm.hwdt cimport arm_halfword_data_transfer
from pyboy_advance.cpu.arm.mul cimport arm_multiply, arm_multiply_long
from pyboy_advance.cpu.arm.psr cimport arm_msr, arm_mrs
from pyboy_advance.cpu.arm.sdt cimport arm_single_data_transfer
from pyboy_advance.cpu.arm.swi cimport arm_software_interrupt
from pyboy_advance.cpu.arm.swp cimport arm_single_data_swap

cdef InstrPattern[12] ARM_PATTERNS

cdef InstrHandler[4096] ARM_LUT

cdef void fill_arm_lut() noexcept

cdef InstrHandler arm_decoder(uint32_t) noexcept



cimport cython

from libc.stdint cimport int64_t
from cpython.time cimport perf_counter_ns

from pyboy_advance.api.screen cimport Screen
from pyboy_advance.app.constants cimport WindowEvent
from pyboy_advance.app.window cimport Window
from pyboy_advance.constants cimport CLOCK_SPEED_HZ, NANOSECONDS_PER_SECOND, EventTrigger, Key
from pyboy_advance.cpu.constants cimport ExceptionVector, BankIndex
from pyboy_advance.cpu.cpu cimport CPU
from pyboy_advance.cpu.arm.decode cimport arm_decoder
from pyboy_advance.cpu.thumb.decode cimport thumb_decoder
from pyboy_advance.interrupt_controller cimport InterruptController
from pyboy_advance.keypad cimport Keypad
from pyboy_advance.memory.dma cimport DMAController
from pyboy_advance.memory.gamepak cimport GamePak
from pyboy_advance.memory.io cimport IO
from pyboy_advance.memory.memory cimport Memory
from pyboy_advance.ppu.constants cimport CYCLES_FRAME
from pyboy_advance.ppu.ppu cimport PPU
from pyboy_advance.scheduler cimport Scheduler
from pyboy_advance.timer cimport Timers

cdef class PyBoyAdvance:
    cdef GamePak gamepak
    cdef Scheduler scheduler
    cdef Memory memory
    cdef InterruptController interrupt_controller
    cdef DMAController dma_controller
    cdef Timers timers
    cdef PPU ppu
    cdef Keypad keypad
    cdef readonly CPU cpu

    cdef readonly Screen screen

    cdef int64_t _time_per_frame
    cdef int64_t _last_frame_time
    cdef int64_t _accumulated_time
    cdef int _frame_overshoot

    cpdef void step(self) noexcept

    cpdef void frame(self, int count=*, bool render=*) noexcept

    @cython.locals(window=Window, event=int)
    cpdef void run(self) noexcept

    cpdef void press_key(self, int) noexcept
    cpdef void release_key(self, int) noexcept

    cpdef void set_emulation_speed(self, float speed)

    cdef void _frame_limiter(self) noexcept



cimport cython

from libc.stdint cimport uint32_t

from pyboy_advance.constants cimport Interrupt, PowerDownMode, EventTrigger
from pyboy_advance.scheduler cimport Scheduler

cdef class InterruptController:
    cdef int WRITE_INTERRUPT_REGISTERS_DELAY
    cdef int UPDATE_IRQ_LINE_DELAY

    cdef Scheduler scheduler
    cdef uint32_t _interrupt_enable
    cdef uint32_t _interrupt_flags
    cdef bint _interrupt_master_enable
    cdef uint32_t _pending_interrupt_enable
    cdef uint32_t _pending_interrupt_flags
    cdef bint _pending_interrupt_master_enable
    cdef bint irq_line
    cdef int power_down_mode

    cdef uint32_t get_interrupt_enable(self) noexcept
    cdef void set_interrupt_enable(self, uint32_t) noexcept
    cdef uint32_t get_interrupt_flags(self) noexcept
    cdef void set_interrupt_flags(self, uint32_t) noexcept
    cdef bint get_interrupt_master_enable(self) noexcept
    cdef void set_interrupt_master_enable(self, bint) noexcept
    cdef void signal(self, int) noexcept
    cdef void _schedule_write_interrupt_registers(self) noexcept

    @cython.locals(new_irq_line=bint)
    cdef void _write_interrupt_registers(self) noexcept

    cdef void _set_irq_line(self) noexcept
    cdef void _reset_irq_line(self) noexcept



from enum import IntEnum


class WindowEvent(IntEnum):
    # fmt: off
    NONE                    = 1

    QUIT                    = 2
    FULLSCREEN              = 3

    PRESS_BUTTON_A          = 4
    PRESS_BUTTON_B          = 5
    PRESS_BUTTON_SELECT     = 6
    PRESS_BUTTON_START      = 7
    PRESS_DPAD_RIGHT        = 8
    PRESS_DPAD_LEFT         = 9
    PRESS_DPAD_UP           = 10
    PRESS_DPAD_DOWN         = 11
    PRESS_SHOULDER_RIGHT    = 12
    PRESS_SHOULDER_LEFT     = 13

    RELEASE_BUTTON_A        = 14
    RELEASE_BUTTON_B        = 15
    RELEASE_BUTTON_SELECT   = 16
    RELEASE_BUTTON_START    = 17
    RELEASE_DPAD_RIGHT      = 18
    RELEASE_DPAD_LEFT       = 19
    RELEASE_DPAD_UP         = 20
    RELEASE_DPAD_DOWN       = 21
    RELEASE_SHOULDER_RIGHT  = 22
    RELEASE_SHOULDER_LEFT   = 23
    # fmt: on



from libc.stdint cimport uint32_t

from pyboy_advance.cpu.cpu cimport CPU
from pyboy_advance.cpu.arm.constants cimport DataTransferOpcode
from pyboy_advance.memory.constants cimport MemoryAccess
from pyboy_advance.utils cimport get_bits, get_bit, add_32

cdef void arm_halfword_data_transfer(CPU, uint32_t) noexcept



# ifndef CYTHON
from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, TypeAlias, TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.cpu.cpu import CPU

from pyboy_advance.cpu.arm.alu import arm_alu
from pyboy_advance.cpu.arm.bdt import arm_block_data_transfer
from pyboy_advance.cpu.arm.branch import arm_branch_exchange, arm_branch
from pyboy_advance.cpu.arm.hwdt import arm_halfword_data_transfer
from pyboy_advance.cpu.arm.mul import arm_multiply, arm_multiply_long
from pyboy_advance.cpu.arm.psr import arm_msr, arm_mrs
from pyboy_advance.cpu.arm.sdt import arm_single_data_transfer
from pyboy_advance.cpu.arm.swi import arm_software_interrupt
from pyboy_advance.cpu.arm.swp import arm_single_data_swap

# fmt: off
InstrHandler: TypeAlias = Callable[["CPU", int], None]

@dataclass
class InstrPattern:
    mask: int
    value: int
    handler: InstrHandler
# fmt: on
# endif


ARM_PATTERNS: list[InstrPattern] = [
    # ----------+-----------------+-----------------+-----------------------------
    #           | Mask            | Value           | Handler
    # ----------+-----------------+-----------------+-----------------------------
    InstrPattern(0b1111_0000_0000, 0b1111_0000_0000, arm_software_interrupt),
    InstrPattern(0b1110_0000_0000, 0b1010_0000_0000, arm_branch),
    InstrPattern(0b1110_0000_0000, 0b1000_0000_0000, arm_block_data_transfer),
    InstrPattern(0b1100_0000_0000, 0b0100_0000_0000, arm_single_data_transfer),
    InstrPattern(0b1111_1111_1111, 0b0001_0010_0001, arm_branch_exchange),
    InstrPattern(0b1111_1011_1111, 0b0001_0000_1001, arm_single_data_swap),
    InstrPattern(0b1111_1000_1111, 0b0000_1000_1001, arm_multiply_long),
    InstrPattern(0b1111_1100_1111, 0b0000_0000_1001, arm_multiply),
    InstrPattern(0b1110_0000_1001, 0b0000_0000_1001, arm_halfword_data_transfer),
    InstrPattern(0b1101_1011_0000, 0b0001_0000_0000, arm_mrs),
    InstrPattern(0b1101_1011_0000, 0b0001_0010_0000, arm_msr),
    InstrPattern(0b1100_0000_0000, 0b0000_0000_0000, arm_alu),
]


# ifndef CYTHON
ARM_LUT: list[InstrHandler | None] = [None] * (2**12)
# endif


def fill_arm_lut():
    for lut_index in range(2**12):
        for pattern in ARM_PATTERNS:
            if (lut_index & pattern.mask) == pattern.value:
                ARM_LUT[lut_index] = pattern.handler
                break


fill_arm_lut()


def arm_decoder(instr: int) -> InstrHandler:
    opcode_hash = ((instr >> 16) & 0xFF0) | ((instr >> 4) & 0xF)
    instruction_handler = ARM_LUT[opcode_hash]

    # ifndef CYTHON
    if not instruction_handler:
        raise ValueError(f"Unknown ARM instruction: {instr:032b}")
    # endif

    return instruction_handler



cpdef enum MemoryAccess:
    NON_SEQUENTIAL = 0
    SEQUENTIAL = 1

cpdef enum MemoryRegion:
    BIOS_START = 0
    BIOS_END = 16383
    BIOS_REGION = 0
    BIOS_MASK = 16383
    BIOS_SIZE = 16384
    EWRAM_START = 33554432
    EWRAM_END = 33816575
    EWRAM_REGION = 2
    EWRAM_MASK = 262143
    EWRAM_SIZE = 262144
    IWRAM_START = 50331648
    IWRAM_END = 50364415
    IWRAM_REGION = 3
    IWRAM_MASK = 32767
    IWRAM_SIZE = 32768
    IO_START = 67108864
    IO_END = 67109887
    IO_REGION = 4
    IO_MASK = 1023
    IO_SIZE = 1024
    PALRAM_START = 83886080
    PALRAM_END = 83887103
    PALRAM_REGION = 5
    PALRAM_MASK = 1023
    PALRAM_SIZE = 1024
    VRAM_START = 100663296
    VRAM_END = 100761599
    VRAM_REGION = 6
    VRAM_MASK_1 = 98303
    VRAM_MASK_2 = 131071
    VRAM_SIZE = 98304
    OAM_START = 117440512
    OAM_END = 117441535
    OAM_REGION = 7
    OAM_MASK = 1023
    OAM_SIZE = 1024
    GAMEPAK_0_START = 134217728
    GAMEPAK_0_END = 167772159
    GAMEPAK_0_REGION_1 = 8
    GAMEPAK_0_REGION_2 = 9
    GAMEPAK_1_START = 167772160
    GAMEPAK_1_END = 201326591
    GAMEPAK_1_REGION_1 = 10
    GAMEPAK_1_REGION_2 = 11
    GAMEPAK_2_START = 201326592
    GAMEPAK_2_END = 234881023
    GAMEPAK_2_REGION_1 = 12
    GAMEPAK_2_REGION_2 = 13
    GAMEPAK_MASK = 33554431
    GAMEPAK_SIZE = 33554432
    GAMEPAK_REGION_START = 8
    GAMEPAK_REGION_END = 13
    SRAM_START = 234881024
    SRAM_END = 234913791
    SRAM_REGION = 14
    SRAM_MASK = 32767
    SRAM_SIZE = 32768

cpdef enum IOAddress:
    REG_DISPCNT = 67108864
    REG_DISPSTAT = 67108868
    REG_VCOUNT = 67108870
    REG_BG0CNT = 67108872
    REG_BG1CNT = 67108874
    REG_BG2CNT = 67108876
    REG_BG3CNT = 67108878
    REG_BG0HOFS = 67108880
    REG_BG0VOFS = 67108882
    REG_BG1HOFS = 67108884
    REG_BG1VOFS = 67108886
    REG_BG2HOFS = 67108888
    REG_BG2VOFS = 67108890
    REG_BG3HOFS = 67108892
    REG_BG3VOFS = 67108894
    REG_WIN0H = 67108928
    REG_WIN1H = 67108930
    REG_WIN0V = 67108932
    REG_WIN1V = 67108934
    REG_WININ = 67108936
    REG_WINOUT = 67108938
    REG_BLDCNT = 67108944
    REG_BLDALPHA = 67108946
    REG_BLDY = 67108948
    REG_SOUNDBIAS = 67109000
    REG_FIFO_A = 67109024
    REG_FIFO_B = 67109028
    REG_DMA0SAD_L = 67109040
    REG_DMA0SAD_H = 67109042
    REG_DMA0DAD_L = 67109044
    REG_DMA0DAD_H = 67109046
    REG_DMA1SAD_L = 67109052
    REG_DMA1SAD_H = 67109054
    REG_DMA1DAD_L = 67109056
    REG_DMA1DAD_H = 67109058
    REG_DMA2SAD_L = 67109064
    REG_DMA2SAD_H = 67109066
    REG_DMA2DAD_L = 67109068
    REG_DMA2DAD_H = 67109070
    REG_DMA3SAD_L = 67109076
    REG_DMA3SAD_H = 67109078
    REG_DMA3DAD_L = 67109080
    REG_DMA3DAD_H = 67109082
    REG_DMA0CNT_L = 67109048
    REG_DMA0CNT_H = 67109050
    REG_DMA1CNT_L = 67109060
    REG_DMA1CNT_H = 67109062
    REG_DMA2CNT_L = 67109072
    REG_DMA2CNT_H = 67109074
    REG_DMA3CNT_L = 67109084
    REG_DMA3CNT_H = 67109086
    REG_TM0CNT_L = 67109120
    REG_TM0CNT_H = 67109122
    REG_TM1CNT_L = 67109124
    REG_TM1CNT_H = 67109126
    REG_TM2CNT_L = 67109128
    REG_TM2CNT_H = 67109130
    REG_TM3CNT_L = 67109132
    REG_TM3CNT_H = 67109134
    REG_KEYINPUT = 67109168
    REG_KEYCNT = 67109170
    REG_IE = 67109376
    REG_IF = 67109378
    REG_WAITCNT = 67109380
    REG_IME = 67109384
    REG_HALTCNT = 67109632

cpdef enum DMAAddressAdjustment:
    INCREMENT = 0
    DECREMENT = 1
    LEAVE_UNCHANGED = 2
    INCREMENT_RELOAD = 3

cpdef enum DMATransferSize:
    HALFWORD = 0
    WORD = 1

cpdef enum DMAStartTiming:
    DMAStartTiming_IMMEDIATELY = 0
    DMAStartTiming_VBLANK = 1
    DMAStartTiming_HBLANK = 2
    SPECIAL = 3




# ifndef CYTHON
from __future__ import annotations

from pyboy_advance.app.constants import WindowEvent
from pyboy_advance.ppu.constants import DISPLAY_WIDTH, DISPLAY_HEIGHT, COLOUR_SIZE
# endif

from ctypes import c_void_p

import sdl2.ext


# fmt: off
DEFAULT_KEY_DOWN_MAP = {
    sdl2.SDLK_a         : WindowEvent.PRESS_BUTTON_A,
    sdl2.SDLK_s         : WindowEvent.PRESS_BUTTON_B,
    sdl2.SDLK_BACKSPACE : WindowEvent.PRESS_BUTTON_SELECT,
    sdl2.SDLK_RETURN    : WindowEvent.PRESS_BUTTON_START,
    sdl2.SDLK_RIGHT     : WindowEvent.PRESS_DPAD_RIGHT,
    sdl2.SDLK_LEFT      : WindowEvent.PRESS_DPAD_LEFT,
    sdl2.SDLK_UP        : WindowEvent.PRESS_DPAD_UP,
    sdl2.SDLK_DOWN      : WindowEvent.PRESS_DPAD_DOWN,
    sdl2.SDLK_w         : WindowEvent.PRESS_SHOULDER_RIGHT,
    sdl2.SDLK_q         : WindowEvent.PRESS_SHOULDER_LEFT,
}

DEFAULT_KEY_UP_MAP = {
    sdl2.SDLK_a         : WindowEvent.RELEASE_BUTTON_A,
    sdl2.SDLK_s         : WindowEvent.RELEASE_BUTTON_B,
    sdl2.SDLK_BACKSPACE : WindowEvent.RELEASE_BUTTON_SELECT,
    sdl2.SDLK_RETURN    : WindowEvent.RELEASE_BUTTON_START,
    sdl2.SDLK_RIGHT     : WindowEvent.RELEASE_DPAD_RIGHT,
    sdl2.SDLK_LEFT      : WindowEvent.RELEASE_DPAD_LEFT,
    sdl2.SDLK_UP        : WindowEvent.RELEASE_DPAD_UP,
    sdl2.SDLK_DOWN      : WindowEvent.RELEASE_DPAD_DOWN,
    sdl2.SDLK_w         : WindowEvent.RELEASE_SHOULDER_RIGHT,
    sdl2.SDLK_q         : WindowEvent.RELEASE_SHOULDER_LEFT,
    sdl2.SDLK_ESCAPE    : WindowEvent.QUIT,
    sdl2.SDLK_F11       : WindowEvent.FULLSCREEN,
}
# fmt: on


class Window:
    def __init__(self):
        self._events = []
        self._key_down_map = DEFAULT_KEY_DOWN_MAP
        self._key_up_map = DEFAULT_KEY_UP_MAP
        self._window_scale = 3

    def __enter__(self):
        if sdl2.SDL_InitSubSystem(sdl2.SDL_INIT_VIDEO | sdl2.SDL_INIT_GAMECONTROLLER) < 0:
            raise ValueError(f"SDL_InitSubSystem failed: {sdl2.SDL_GetError().decode()}")

        self._sdl_window = sdl2.SDL_CreateWindow(
            b"PyBoy Advance",
            sdl2.SDL_WINDOWPOS_CENTERED,
            sdl2.SDL_WINDOWPOS_CENTERED,
            DISPLAY_WIDTH * self._window_scale,
            DISPLAY_HEIGHT * self._window_scale,
            sdl2.SDL_WINDOW_RESIZABLE,
        )

        self._sdl_renderer = sdl2.SDL_CreateRenderer(
            self._sdl_window, -1, sdl2.SDL_RENDERER_ACCELERATED
        )

        sdl2.SDL_RenderSetLogicalSize(self._sdl_renderer, DISPLAY_WIDTH, DISPLAY_HEIGHT)

        self._sdl_texture_buffer = sdl2.SDL_CreateTexture(
            self._sdl_renderer,
            sdl2.SDL_PIXELFORMAT_BGR555,
            sdl2.SDL_TEXTUREACCESS_STREAMING,
            DISPLAY_WIDTH,
            DISPLAY_HEIGHT,
        )

        sdl2.SDL_ShowWindow(self._sdl_window)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sdl2.SDL_DestroyWindow(self._sdl_window)
        sdl2.SDL_QuitSubSystem(sdl2.SDL_INIT_VIDEO | sdl2.SDL_INIT_GAMECONTROLLER)

    @property
    def fullscreen(self) -> bool:
        flags = sdl2.SDL_GetWindowFlags(self._sdl_window)
        return flags & sdl2.SDL_WINDOW_FULLSCREEN_DESKTOP

    @fullscreen.setter
    def fullscreen(self, value: bool):
        sdl2.SDL_SetWindowFullscreen(
            self._sdl_window,
            sdl2.SDL_WINDOW_FULLSCREEN_DESKTOP if value else 0,
        )

    def get_events(self) -> list[WindowEvent]:
        self._events.clear()
        for event in sdl2.ext.get_events():
            if event.type == sdl2.SDL_QUIT:
                self._events.append(WindowEvent.QUIT)
            elif event.type == sdl2.SDL_KEYDOWN:
                self._events.append(self._key_down_map.get(event.key.keysym.sym, WindowEvent.NONE))
            elif event.type == sdl2.SDL_KEYUP:
                self._events.append(self._key_up_map.get(event.key.keysym.sym, WindowEvent.NONE))
        return self._events

    def render(self, frame_buffer_ptr: c_void_p):
        sdl2.SDL_UpdateTexture(
            self._sdl_texture_buffer,
            None,
            frame_buffer_ptr,
            DISPLAY_WIDTH * COLOUR_SIZE,
        )
        sdl2.SDL_RenderClear(self._sdl_renderer)
        sdl2.SDL_RenderCopy(self._sdl_renderer, self._sdl_texture_buffer, None, None)
        sdl2.SDL_RenderPresent(self._sdl_renderer)



cimport cython

from libc.stdint cimport uint32_t

from pyboy_advance.cpu.cpu cimport CPU
from pyboy_advance.cpu.arm.constants cimport ALUOpcode
from pyboy_advance.utils cimport get_bits, get_bit, ror_32, sign_32

cdef void arm_alu(CPU, uint32_t) noexcept

cdef void arm_alu(CPU, uint32_t) noexcept

cdef void arm_alu_and(CPU, uint32_t, uint32_t, uint32_t, bint, bint) noexcept

cdef uint32_t arm_alu_and_impl(CPU, uint32_t, uint32_t, uint32_t, bint, bint) noexcept

cdef void arm_alu_eor(CPU, uint32_t, uint32_t, uint32_t, bint, bint) noexcept

cdef uint32_t arm_alu_eor_impl(CPU, uint32_t, uint32_t, uint32_t, bint, bint) noexcept

cdef void arm_alu_sub(CPU, uint32_t, uint32_t, uint32_t, bint) noexcept

@cython.locals(mask=uint32_t)
cdef uint32_t arm_alu_sub_impl(CPU, uint32_t, uint32_t, uint32_t, bint) noexcept

cdef void arm_alu_rsb(CPU, uint32_t, uint32_t, uint32_t, bint) noexcept

cdef void arm_alu_add(CPU, uint32_t, uint32_t, uint32_t, bint) noexcept

@cython.locals(mask=uint32_t)
cdef uint32_t arm_alu_add_impl(CPU, uint32_t, uint32_t, uint32_t, bint) noexcept

@cython.locals(mask=uint32_t, carry_1=bint, carry_2=bint)
cdef void arm_alu_adc(CPU, uint32_t, uint32_t, uint32_t, bint) noexcept

@cython.locals(mask=uint32_t, borrow_1=bint, borrow_2=bint)
cdef void arm_alu_sbc(CPU, uint32_t, uint32_t, uint32_t, bint) noexcept

cdef void arm_alu_rsc(CPU, uint32_t, uint32_t, uint32_t, bint) noexcept

cdef void arm_alu_tst(CPU, uint32_t, uint32_t, bint) noexcept

cdef void arm_alu_teq(CPU, uint32_t, uint32_t, bint) noexcept

cdef void arm_alu_cmp(CPU, uint32_t, uint32_t) noexcept

cdef void arm_alu_cmn(CPU, uint32_t, uint32_t) noexcept

cdef void arm_alu_orr(CPU, uint32_t, uint32_t, uint32_t, bint, bint) noexcept

cdef void arm_alu_mov(CPU, uint32_t, uint32_t, bint, bint) noexcept

cdef void arm_alu_bic(CPU, uint32_t, uint32_t, uint32_t, bint, bint) noexcept

@cython.locals(mask=uint32_t)
cdef void arm_alu_mvn(CPU, uint32_t, uint32_t, bint, bint) noexcept



class DMAControlRegister:

    def __init__(self):
        self.reg = 0

    def get_dst_address_adjustment(self) -> DMAAddressAdjustment | int:
        return get_bits(self.reg, 5, 6)

    def get_src_address_adjustment(self) -> DMAAddressAdjustment | int:
        return get_bits(self.reg, 7, 8)

    def get_repeat(self) -> bint:
        return get_bit(self.reg, 9)

    def get_transfer_size(self) -> DMATransferSize | int:
        return get_bit(self.reg, 10)

    def get_start_timing(self) -> DMAStartTiming | int:
        return get_bits(self.reg, 12, 13)

    def get_irq_when_done(self) -> bint:
        return get_bit(self.reg, 14)

    def get_transfer_enabled(self) -> bint:
        return get_bit(self.reg, 15)

    def set_transfer_enabled(self, enable: bint):
        self.reg = set_bit(self.reg, 15, enable)

class DMAChannel:

    def __init__(self, channel_id: int, scheduler: Scheduler, memory: Memory):
        self.SRC_MASK = [134217727, 268435455, 268435455, 268435455]
        self.DST_MASK = [134217727, 134217727, 134217727, 268435455]
        self.COUNT_MASK = [16383, 16383, 16383, 65535]
        self.INTERRUPT = [Interrupt.DMA_0, Interrupt.DMA_1, Interrupt.DMA_2, Interrupt.DMA_3]
        self.TRANSFER_DELAY = 2
        self.channel_id = channel_id
        self.scheduler = scheduler
        self.memory = memory
        self.fifo = False
        self.pending = False
        self._control_reg = DMAControlRegister()
        self._count = 0
        self._src = 0
        self._dst = 0
        self._internal_count = 0
        self._internal_src = 0
        self._internal_dst = 0
        self._event = None

    def get_control_reg(self) -> int:
        return self._control_reg.reg

    def set_control_reg(self, value: int) -> None:
        old_enable = self._control_reg.get_transfer_enabled()
        self._control_reg.reg = value
        if not old_enable and self._control_reg.get_transfer_enabled():
            self.fifo = self._control_reg.get_start_timing() == DMAStartTiming.SPECIAL and (self.channel_id == 1 or self.channel_id == 2) and (self._dst == IOAddress.REG_FIFO_A or self._dst == IOAddress.REG_FIFO_B)
            self._internal_src = self._src
            self._internal_dst = self._dst
            self._internal_count = self._count
            if self._internal_count == 0:
                self._internal_count = self.COUNT_MASK[self.channel_id] + 1
            if self._control_reg.get_start_timing() == DMAStartTiming.DMAStartTiming_IMMEDIATELY:
                self._event = self.scheduler.schedule(self.activate, self.TRANSFER_DELAY, EventTrigger.IMMEDIATELY)
            elif self._control_reg.get_start_timing() == DMAStartTiming.DMAStartTiming_VBLANK:
                self._event = self.scheduler.schedule(self.activate, self.TRANSFER_DELAY, EventTrigger.EventTrigger_VBLANK)
            elif self._control_reg.get_start_timing() == DMAStartTiming.DMAStartTiming_HBLANK:
                self._event = self.scheduler.schedule(self.activate, self.TRANSFER_DELAY, EventTrigger.EventTrigger_HBLANK)
        elif old_enable and (not self._control_reg.get_transfer_enabled()):
            if self._event:
                self._event.cancelled = False
                self._event = None
            self.pending = False

    def get_count(self) -> int:
        return self._count

    def set_count(self, value: int) -> None:
        self._count = value & self.COUNT_MASK[self.channel_id]

    def get_src_address(self) -> int:
        return self._src

    def set_src_address(self, value: int):
        self._src = value & self.SRC_MASK[self.channel_id]

    def get_dst_address(self) -> int:
        return self._dst

    def set_dst_address(self, value: int):
        self._dst = value & self.DST_MASK[self.channel_id]

    def transfer(self):
        if not self.pending:
            return
        transfer_size = self._control_reg.get_transfer_size()
        transfer_size_bytes = 4 if transfer_size == DMATransferSize.WORD else 2
        address_alignment = ~3 if transfer_size == DMATransferSize.WORD else ~1
        self._internal_src &= address_alignment
        self._internal_dst &= address_alignment
        src_adj = self._control_reg.get_src_address_adjustment()
        if src_adj == DMAAddressAdjustment.INCREMENT:
            src_step = transfer_size_bytes
        elif src_adj == DMAAddressAdjustment.DECREMENT:
            src_step = -transfer_size_bytes
        elif src_adj == DMAAddressAdjustment.LEAVE_UNCHANGED:
            src_step = 0
        dst_adj = self._control_reg.get_dst_address_adjustment()
        if self.fifo:
            dst_step = 0
        elif dst_adj == DMAAddressAdjustment.INCREMENT:
            dst_step = transfer_size_bytes
        elif dst_adj == DMAAddressAdjustment.DECREMENT:
            dst_step = -transfer_size_bytes
        elif dst_adj == DMAAddressAdjustment.LEAVE_UNCHANGED:
            dst_step = 0
        elif dst_adj == DMAAddressAdjustment.INCREMENT_RELOAD:
            dst_step = transfer_size_bytes
        access = MemoryAccess.NON_SEQUENTIAL
        for _ in range(self._internal_count):
            if transfer_size == DMATransferSize.WORD:
                value = self.memory.read_32(self._internal_src, access)
                self.memory.write_32(self._internal_dst, value, access)
            else:
                value = self.memory.read_16(self._internal_src, access)
                self.memory.write_16(self._internal_dst, value, access)
            self._internal_src += src_step
            self._internal_dst += dst_step
            access = MemoryAccess.SEQUENTIAL
        if self._control_reg.get_irq_when_done():
            self.memory.io.interrupt_controller.signal(self.INTERRUPT[self.channel_id])
        self.pending = False
        if self._control_reg.get_repeat():
            if dst_adj == DMAAddressAdjustment.INCREMENT_RELOAD:
                self._internal_dst = self._dst
            if self._control_reg.get_start_timing() == DMAStartTiming.DMAStartTiming_VBLANK:
                self._event = self.scheduler.schedule(self.activate, self.TRANSFER_DELAY, EventTrigger.EventTrigger_VBLANK)
            elif self._control_reg.get_start_timing() == DMAStartTiming.DMAStartTiming_HBLANK:
                self._event = self.scheduler.schedule(self.activate, self.TRANSFER_DELAY, EventTrigger.EventTrigger_HBLANK)
        else:
            self._control_reg.set_transfer_enabled(False)

    def activate(self):
        if self._control_reg.get_transfer_enabled():
            self.pending = True

class DMAController:

    def __init__(self, scheduler: Scheduler, memory: Memory):
        self.channel_0 = DMAChannel(0, scheduler, memory)
        self.channel_1 = DMAChannel(1, scheduler, memory)
        self.channel_2 = DMAChannel(2, scheduler, memory)
        self.channel_3 = DMAChannel(3, scheduler, memory)

    def get_active(self) -> bool:
        return self.channel_0.pending or self.channel_1.pending or self.channel_2.pending or self.channel_3.pending

    def perform_transfers(self):
        self.channel_0.transfer()
        self.channel_1.transfer()
        self.channel_2.transfer()
        self.channel_3.transfer()


import sys
from pathlib import Path

sys.path.insert(0, str(Path(__file__).parent))

from build_utils.property_preprocessor import preprocess_properties
from build_utils.constants_preprocessor import preprocess_constants
from build_utils.cython_guard_preprocessor import preprocess_cython_guards

import multiprocessing
import os
import platform
import shutil
from multiprocessing import cpu_count
import numpy as np

from setuptools import Extension, setup

CYTHON = platform.python_implementation() == "CPython"
ROOT_DIR = "pyboy_advance"
DEBUG = bool(os.getenv("GITHUB_ACTIONS"))

if not CYTHON:
    setup()
    exit(0)

from Cython.Build import cythonize  # noqa
from Cython.Compiler import DebugFlags, Errors  # noqa
from Cython.Distutils import build_ext as _build_ext  # noqa


def patched_error(position, message):
    if message == "Python object cannot be passed as a varargs parameter":
        # We ignore this error to pass PyObject* to logging
        return
    else:
        # print("Errors.error:", repr(position), repr(message)) ###
        if position is None:
            raise Errors.InternalError(message)
        err = Errors.CompileError(position, message)
        if DebugFlags.debug_exception_on_error:
            raise Exception(err)  # debug
        Errors.report_error(err)
        return err


Errors.error = patched_error


class BuildExt(_build_ext):
    def initialize_options(self):
        super().initialize_options()
        # This is only for Cythonize. See finalize_options.
        if sys.platform == "win32":
            thread_count = 0  # Disables multiprocessing
        else:
            thread_count = cpu_count()

        # Fixing issue with nthreads in Cython
        if ((3, 8) <= sys.version_info[:2] and sys.platform == "darwin") or (
            (3, 14) <= sys.version_info[:2] and sys.platform == "linux"
        ):
            multiprocessing.set_start_method("fork", force=True)

        cflags = ["-O3"]
        if not DEBUG:
            cflags.append("-DCYTHON_WITHOUT_ASSERTIONS")
        # NOTE: For performance. Check if other platforms need an equivalent change.
        if sys.platform == "darwin":
            cflags.append(
                "-DCYTHON_INLINE=inline __attribute__ ((__unused__)) __attribute__((always_inline))"
            )

        py_pxd_files = list(prep_pxd_py_files())

        preprocessed_files = []
        for src_file in py_pxd_files:
            build_file = src_file.replace(".py", ".build.py")
            shutil.copy(src_file, build_file)
            preprocessed_files.append(build_file)

        preprocessed_files = preprocess_cython_guards(preprocessed_files)
        preprocessed_files = preprocess_properties(preprocessed_files)
        preprocessed_files = preprocess_constants(preprocessed_files)

        cythonize_files = map(
            lambda src: Extension(
                src.split(".")[0].replace(os.sep, "."),
                [src],
                extra_compile_args=cflags,
                extra_link_args=[] if DEBUG else ["-s", "-w"],
                include_dirs=[np.get_include()],
            ),
            preprocessed_files,
        )
        self.distribution.ext_modules = cythonize(
            [*cythonize_files],
            nthreads=thread_count,
            annotate=True,
            gdb_debug=False,
            language_level=3,
            compiler_directives={
                "annotation_typing": False,
                "boundscheck": False,
                "cdivision": True,
                "cdivision_warnings": False,
                "infer_types": True,
                "initializedcheck": False,
                "nonecheck": False,
                "overflowcheck": False,
                "wraparound": False,
                "legacy_implicit_noexcept": True,
            },
        )

    def finalize_options(self):
        # Use '-j' on the commandline to specify parallelism. Otherwise auto-detect.
        if getattr(self, "parallel", None) is None:
            self.parallel = cpu_count()
        super().finalize_options()


def prep_pxd_py_files():
    ignore_py_files = ["__main__.py", "__init__.py", "conftest.py"]
    # Cython doesn't trigger a recompile on .py files, where only the .pxd file has changed. So we fix this here.
    # We also yield the py_files that have a .pxd file, as we feed these into the cythonize call.
    for root, dirs, files in os.walk(ROOT_DIR):
        for f in files:
            base, ext = f.split(".", 1)
            if ext == "py" and f not in ignore_py_files:
                yield os.path.join(root, f)
            if ext == "pxd":
                py_file = os.path.join(root, base) + ".py"
                if os.path.isfile(py_file):
                    if os.path.getmtime(os.path.join(root, f)) > os.path.getmtime(py_file):
                        os.utime(py_file)


setup(
    cmdclass={
        "build_ext": BuildExt,
    },
    ext_modules=[Extension("", [""])],  # Added to trigger a binary wheel
)



# ifndef CYTHON
from time import perf_counter_ns

from pyboy_advance.api.screen import Screen
from pyboy_advance.app.constants import WindowEvent
from pyboy_advance.app.window import Window
from pyboy_advance.constants import CLOCK_SPEED_HZ, NANOSECONDS_PER_SECOND, EventTrigger, Key
from pyboy_advance.cpu.constants import ExceptionVector, BankIndex
from pyboy_advance.cpu.cpu import CPU
from pyboy_advance.cpu.arm.decode import arm_decoder
from pyboy_advance.cpu.thumb.decode import thumb_decoder
from pyboy_advance.interrupt_controller import InterruptController
from pyboy_advance.keypad import Keypad
from pyboy_advance.memory.dma import DMAController
from pyboy_advance.memory.gamepak import GamePak
from pyboy_advance.memory.io import IO
from pyboy_advance.memory.memory import Memory
from pyboy_advance.ppu.constants import CYCLES_FRAME
from pyboy_advance.ppu.ppu import PPU
from pyboy_advance.scheduler import Scheduler
from pyboy_advance.timer import Timers
# endif

import os

from time import sleep


class PyBoyAdvance:
    def __init__(
        self,
        rom: GamePak | str | os.PathLike,
        bios: str | os.PathLike | None = None,
        skip_bios: bool = False,
        emulation_speed: float = 0,
    ):
        self.gamepak = GamePak.from_file(rom) if isinstance(rom, (str, os.PathLike)) else rom

        if bios:
            with open(bios, "rb") as bios_file:
                bios_data = bytearray(bios_file.read())
        else:
            bios_data = b""

        self.scheduler = Scheduler()
        self.memory = Memory(self.scheduler, self.gamepak, bios_data)

        self.interrupt_controller = InterruptController(self.scheduler)
        self.dma_controller = DMAController(self.scheduler, self.memory)
        self.timers = Timers(self.scheduler, self.interrupt_controller)
        self.ppu = PPU(self.scheduler, self.interrupt_controller)
        self.keypad = Keypad(self.interrupt_controller)
        self.memory.io = IO(
            self.memory,
            self.interrupt_controller,
            self.dma_controller,
            self.timers,
            self.ppu,
            self.keypad,
        )

        self.cpu = CPU(self.scheduler, self.memory)
        self.cpu.arm_decoder = arm_decoder
        self.cpu.thumb_decoder = thumb_decoder

        self._time_per_frame = 0
        self._last_frame_time = 0
        self._accumulated_time = 0

        self.set_emulation_speed(emulation_speed)
        self.scheduler.schedule(self._frame_limiter, CYCLES_FRAME, EventTrigger.IMMEDIATELY)

        self._frame_overshoot = 0

        if skip_bios:
            self.cpu.regs.banked_sp[int(BankIndex.BANK_SYSTEM_USER)] = 0x03007F00
            self.cpu.regs.banked_sp[int(BankIndex.BANK_FIQ)] = 0x03007F00
            self.cpu.regs.banked_sp[int(BankIndex.BANK_IRQ)] = 0x03007FA0
            self.cpu.regs.banked_sp[int(BankIndex.BANK_SWI)] = 0x03007FE0
            self.cpu.regs.banked_sp[int(BankIndex.BANK_ABORT)] = 0x03007F00
            self.cpu.regs.banked_sp[int(BankIndex.BANK_UNDEFINED)] = 0x03007F00
            self.cpu.regs.sp = 0x03007F00
            self.cpu.regs.pc = 0x08000000
            self.cpu.flush_pipeline()
        else:
            self.cpu.interrupt(ExceptionVector.RESET)

        self.screen = Screen(self.ppu)

    def step(self):
        if self.dma_controller.active:
            self.dma_controller.perform_transfers()
        else:
            self.cpu.step()

    def frame(self, count: int = 1, render: bool = True):
        self.ppu.rendering_enabled = False

        for frame_num in range(count):
            if frame_num == count - 1 and render:
                self.ppu.rendering_enabled = True

            end_time = self.scheduler.cycles + CYCLES_FRAME - self._frame_overshoot
            while self.scheduler.cycles < end_time:
                self.step()
            self._frame_overshoot = self.scheduler.cycles - end_time

    def run(self):
        with Window() as window:
            running = True
            while running:
                self.frame()

                for event in window.get_events():
                    if event == WindowEvent.NONE:
                        continue
                    elif event == WindowEvent.QUIT:
                        running = False
                    elif event == WindowEvent.FULLSCREEN:
                        window.fullscreen = not window.fullscreen
                    else:
                        self.keypad.process_window_event(event)

                window.render(self.ppu.frame_buffer_ptr)

    def press_key(self, key: Key):
        self.keypad.press_key(key)

    def release_key(self, key: Key):
        self.keypad.release_key(key)

    def set_emulation_speed(self, speed: float):
        if speed < 0:
            raise ValueError("Emulation speed must be 0 or greater")

        self._time_per_frame = (
            int((1 / (speed * CLOCK_SPEED_HZ / CYCLES_FRAME)) * NANOSECONDS_PER_SECOND)
            if speed > 0
            else 0
        )

        self._last_frame_time = perf_counter_ns()
        self._accumulated_time = 0

    def _frame_limiter(self):
        self.scheduler.schedule(self._frame_limiter, CYCLES_FRAME, EventTrigger.IMMEDIATELY)

        if self._time_per_frame == 0:
            return

        now = perf_counter_ns()
        self._accumulated_time += now - self._last_frame_time
        self._last_frame_time = now

        if self._accumulated_time < self._time_per_frame:
            delay = (self._time_per_frame - self._accumulated_time) / float(NANOSECONDS_PER_SECOND)
            sleep(delay)

        self._accumulated_time -= self._time_per_frame



# ifndef CYTHON
from __future__ import annotations

from array import array

from pyboy_advance.constants import Interrupt, EventTrigger
from pyboy_advance.interrupt_controller import InterruptController
from pyboy_advance.ppu.constants import *
from pyboy_advance.ppu.memory import VideoMemory
from pyboy_advance.ppu.registers import (
    DisplayControlRegister,
    DisplayStatusRegister,
    BackgroundControlRegister,
    WindowControlRegister,
    BlendControlRegister,
    BlendAlphaRegister,
    BlendBrightnessRegister,
)
from pyboy_advance.scheduler import Scheduler
from pyboy_advance.utils import (
    get_bit,
    get_bits,
    bint,
    sign_32,
    add_32,
    extend_sign_9,
)
# endif

from ctypes import c_void_p


class PPU:
    """
    Picture Processing Unit.
    """

    def __init__(self, scheduler: Scheduler, interrupt_controller: InterruptController):
        self.scheduler = scheduler
        self.interrupt_controller = interrupt_controller

        self.rendering_enabled = True

        self.display_control = DisplayControlRegister()
        self.display_status = DisplayStatusRegister()
        self.vcount = 0

        self.bg_control_0 = BackgroundControlRegister()
        self.bg_control_1 = BackgroundControlRegister()
        self.bg_control_2 = BackgroundControlRegister()
        self.bg_control_3 = BackgroundControlRegister()

        self.bg_offset_h = array("I", [0] * NUM_BACKGROUNDS)
        self.bg_offset_v = array("I", [0] * NUM_BACKGROUNDS)

        self.window_v_min = array("I", [0] * NUM_PRIMARY_WINDOWS)
        self.window_v_max = array("I", [0] * NUM_PRIMARY_WINDOWS)
        self.window_h_min = array("I", [0] * NUM_PRIMARY_WINDOWS)
        self.window_h_max = array("I", [0] * NUM_PRIMARY_WINDOWS)

        self.window_control_0 = WindowControlRegister()
        self.window_control_1 = WindowControlRegister()
        self.window_control_obj = WindowControlRegister()
        self.window_control_out = WindowControlRegister()

        self.window_mask_0 = array("H", [0] * DISPLAY_WIDTH)
        self.window_mask_1 = array("H", [0] * DISPLAY_WIDTH)
        self.window_mask_obj = array("H", [0] * DISPLAY_WIDTH)

        self.blend_control = BlendControlRegister()
        self.blend_alpha = BlendAlphaRegister()
        self.blend_brightness = BlendBrightnessRegister()

        self.blend_mask_obj = [[False] * DISPLAY_WIDTH for _ in range(NUM_PRIORITIES)]

        self.memory = VideoMemory(self.display_control)

        self.bg_scanline = [array("H", [0] * DISPLAY_WIDTH) for _ in range(NUM_BACKGROUNDS)]
        self.obj_scanline = [array("H", [0] * DISPLAY_WIDTH) for _ in range(NUM_PRIORITIES)]

        self.scanline = array("H", [0] * DISPLAY_WIDTH)
        self.scanline_top_colour = array("H", [0] * DISPLAY_WIDTH)
        self.scanline_top_layer = array("I", [LayerType.BACKDROP] * DISPLAY_WIDTH)

        self.front_buffer = array("H", [0] * (DISPLAY_WIDTH * DISPLAY_HEIGHT))
        self.back_buffer = array("H", [0] * (DISPLAY_WIDTH * DISPLAY_HEIGHT))
        self.front_buffer_ptr = c_void_p(self.front_buffer.buffer_info()[0])
        self.back_buffer_ptr = c_void_p(self.back_buffer.buffer_info()[0])

        self.scheduler.schedule(
            self.hblank_start,
            CYCLES_HDRAW,
            EventTrigger.IMMEDIATELY,
        )

    @property
    def frame_buffer_ptr(self) -> c_void_p:
        return self.front_buffer_ptr

    def hblank_start(self):
        self.display_status.hblank_status = True

        if self.vcount < DISPLAY_HEIGHT:
            if self.rendering_enabled:
                self._init_layers()
                self._calc_window_masks()

                if not self.display_control.force_blank:
                    self._render_backgrounds()
                    self._render_objects()

                self._merge_layers()

            self.scheduler.trigger(EventTrigger.HBLANK)

        if self.display_status.hblank_irq:
            self.interrupt_controller.signal(Interrupt.HBLANK)

        self.scheduler.schedule(
            self.hblank_end,
            CYCLES_HBLANK,
            EventTrigger.IMMEDIATELY,
        )

    def hblank_end(self):
        self.vcount += 1
        if self.vcount >= DISPLAY_FULL_HEIGHT:
            self.vcount = 0

        self.display_status.hblank_status = False
        self.display_status.vblank_status = self.vcount >= DISPLAY_HEIGHT
        self.display_status.vcount_trigger_status = (
            self.vcount == self.display_status.vcount_trigger_value
        )

        if self.vcount == DISPLAY_HEIGHT:
            # Swap front and back buffers
            # (done at this point to avoid tearing, now that the back buffer is complete)
            self.front_buffer, self.back_buffer = self.back_buffer, self.front_buffer
            self.front_buffer_ptr, self.back_buffer_ptr = (
                self.back_buffer_ptr,
                self.front_buffer_ptr,
            )

            if self.display_status.vblank_irq:
                self.interrupt_controller.signal(Interrupt.VBLANK)

            self.scheduler.trigger(EventTrigger.VBLANK)

        if self.display_status.vcount_irq and self.display_status.vcount_trigger_status:
            self.interrupt_controller.signal(Interrupt.VCOUNT)

        self.scheduler.schedule(
            self.hblank_start,
            CYCLES_HDRAW,
            EventTrigger.IMMEDIATELY,
        )

    def _init_layers(self):
        backdrop_colour = 0 if self.display_control.force_blank else self.memory.read_16_palram(0)
        for x in range(DISPLAY_WIDTH):
            self.scanline[x] = backdrop_colour
            self.scanline_top_colour[x] = backdrop_colour
            self.scanline_top_layer[x] = LayerType.BACKDROP

        for scanline in self.bg_scanline:
            for x in range(DISPLAY_WIDTH):
                scanline[x] = TRANSPARENT_COLOUR

        for scanline in self.obj_scanline:
            for x in range(DISPLAY_WIDTH):
                scanline[x] = TRANSPARENT_COLOUR

        for blend_mask in self.blend_mask_obj:
            for x in range(DISPLAY_WIDTH):
                blend_mask[x] = False

    def _render_backgrounds(self):
        video_mode = self.display_control.video_mode

        if video_mode == VideoMode.MODE_0:
            for bg_num in range(NUM_BACKGROUNDS):
                self._render_background_text(bg_num)

        elif video_mode == VideoMode.MODE_1:
            self._render_background_affine(2)
            self._render_background_text(1)
            self._render_background_text(0)

        elif video_mode == VideoMode.MODE_2:
            self._render_background_affine(3)
            self._render_background_affine(2)

        elif video_mode == VideoMode.MODE_3:
            self._render_background_bitmap(False, False)

        elif video_mode == VideoMode.MODE_4:
            self._render_background_bitmap(True, False)

        elif video_mode == VideoMode.MODE_5:
            self._render_background_bitmap(False, True)

    def _render_background_text(self, bg_num: int):
        if not self.display_control.display_bg(bg_num):
            return

        bg_control = self._get_bg_control(bg_num)

        rel_y = self.vcount + self.bg_offset_v[bg_num]
        tile_y = rel_y // TILE_HEIGHT
        pixel_y = rel_y % TILE_HEIGHT

        for x in range(DISPLAY_WIDTH):
            rel_x = x + self.bg_offset_h[bg_num]
            tile_x = rel_x // TILE_WIDTH
            pixel_x = rel_x % TILE_WIDTH

            # Larger size backgrounds are accessed as (up to) 4 separate maps
            # that are laid out as contiguous blocks in VRAM, pictured below.
            # (each tile map block is 32 x 32 tiles in size)
            #
            # +-----------+-----------+-----------+-----------+
            # |  32 x 32  |  64 x 32  |  32 x 64  |  64 x 64  |
            # +-----------+-----------+-----------+-----------+
            # |    [0]    |   [0][1]  |    [0]    |   [0][1]  |
            # |           |           |    [1]    |   [2][3]  |
            # +-----------+-----------+-----------+-----------+
            #
            # Hence, we must first find the tile map block we are currently in
            # and then find the tile map entry inside that block.

            tile_map_block = bg_control.tile_map_block
            if bg_control.size == 0b01:  # 64 x 32
                if tile_x & TILE_MAP_BLOCK_WIDTH:
                    tile_map_block += 1
            elif bg_control.size == 0b10:  # 32 x 64
                if tile_y & TILE_MAP_BLOCK_HEIGHT:
                    tile_map_block += 1
            elif bg_control.size == 0b11:  # 64 x 64
                if tile_x & TILE_MAP_BLOCK_WIDTH:
                    tile_map_block += 1
                if tile_y & TILE_MAP_BLOCK_HEIGHT:
                    tile_map_block += 2

            tile_map_entry_address = (tile_y % TILE_MAP_BLOCK_HEIGHT) * TILE_MAP_BLOCK_WIDTH
            tile_map_entry_address += tile_x % TILE_MAP_BLOCK_WIDTH
            tile_map_entry_address *= TILE_MAP_ENTRY_SIZE
            tile_map_entry_address += tile_map_block * TILE_MAP_BLOCK_SIZE

            tile_map_entry = self.memory.read_16_vram(tile_map_entry_address)
            tile_num = get_bits(tile_map_entry, 0, 9)
            tile_flip_h = get_bit(tile_map_entry, 10)
            tile_flip_v = get_bit(tile_map_entry, 11)
            palette_num = get_bits(tile_map_entry, 12, 15)

            pixel_x_flipped = TILE_WIDTH - pixel_x - 1 if tile_flip_h else pixel_x
            pixel_y_flipped = TILE_HEIGHT - pixel_y - 1 if tile_flip_v else pixel_y

            palette_index = self._get_palette_index(
                bg_control.tile_set_block * TILE_SET_BLOCK_SIZE,
                tile_num,
                pixel_x_flipped,
                pixel_y_flipped,
                bg_control.colour_256,
                palette_num,
            )

            if palette_index:
                colour = self.memory.read_16_palram(palette_index * COLOUR_SIZE)
                self.bg_scanline[bg_num][x] = colour

    def _render_background_affine(self, bg_num: int):
        if not self.display_control.display_bg(bg_num):
            return

    def _render_background_bitmap(self, paletted, small):
        if not self.display_control.display_bg(2):
            return

        page_offset = VRAM_PAGE_OFFSET if self.display_control.page_select else 0

        display_width = SMALL_DISPLAY_WIDTH if small else DISPLAY_WIDTH
        display_height = SMALL_DISPLAY_HEIGHT if small else DISPLAY_HEIGHT

        if paletted:
            for x in range(display_width):
                palette_index = self.memory.read_8_vram(
                    page_offset + display_width * self.vcount + x
                )
                if palette_index:
                    colour = self.memory.read_16_palram(palette_index * COLOUR_SIZE)
                    self.bg_scanline[2][x] = colour
        else:
            for x in range(display_width):
                vram_index = display_width * self.vcount + x
                colour = self.memory.read_16_vram(vram_index * COLOUR_SIZE)
                self.bg_scanline[2][x] = colour

    def _render_objects(self):
        if not self.display_control.display_obj:
            return

        for obj_index in range(MAX_NUM_OBJECTS - 1, -1, -1):
            obj_address = obj_index * OAM_ENTRY_SIZE
            self._render_object(
                Object(
                    self.memory.read_16_oam(obj_address),
                    self.memory.read_16_oam(obj_address + 2),
                    self.memory.read_16_oam(obj_address + 4),
                )
            )

    def _render_object(self, obj: Object):
        if obj.disabled:
            return

        # Because the BG section of VRAM gets extended in bitmapped modes (hence
        # making the OBJ section smaller), attempts to use tiles 0-511 are ignored
        video_mode = self.display_control.video_mode
        bitmapped = video_mode in [VideoMode.MODE_3, VideoMode.MODE_4, VideoMode.MODE_5]
        if bitmapped and obj.tile_index < 512:
            return

        obj_x = extend_sign_9(obj.x)
        obj_y = obj.y
        obj_w, obj_h = obj.size
        obj_flip_h, obj_flip_v = obj.flip_horizontal, obj.flip_vertical

        if obj_y + obj_h > 256:
            obj_y -= 256

        if self.vcount < obj_y or self.vcount >= obj_y + obj_h:
            return

        tile_row_len = obj_w // TILE_WIDTH if self.display_control.obj_vram_dimension else 32

        offset_y = self.vcount - obj_y
        rel_y = obj_y - offset_y - 1 if obj_flip_v else offset_y

        for offset_x in range(0, obj_w):
            win_x = add_32(obj_x, offset_x)
            if sign_32(win_x) or win_x >= DISPLAY_WIDTH:
                continue

            rel_x = obj_w - offset_x - 1 if obj_flip_h else offset_x

            tile_x = rel_x // TILE_WIDTH
            tile_y = rel_y // TILE_HEIGHT
            tile_index = obj.tile_index + tile_y * tile_row_len + tile_x

            pixel_x = rel_x % TILE_WIDTH
            pixel_y = rel_y % TILE_HEIGHT

            palette_index = self._get_palette_index(
                OBJ_TILE_SET_OFFSET,
                tile_index,
                pixel_x,
                pixel_y,
                obj.colour_256,
                obj.palette_num,
            )

            if palette_index:
                if obj.mode == ObjectMode.WINDOW:
                    self.window_mask_obj[win_x] = 1
                else:
                    colour = self.memory.read_16_palram(
                        OBJ_PALETTE_OFFSET + palette_index * COLOUR_SIZE
                    )
                    self.obj_scanline[obj.priority][win_x] = colour
                    self.blend_mask_obj[obj.priority][win_x] = obj.mode == ObjectMode.BLEND

    def _merge_layers(self):
        for priority in range(NUM_PRIORITIES - 1, -1, -1):
            for bg_num in range(NUM_BACKGROUNDS - 1, -1, -1):
                if (
                    self.display_control.display_bg(bg_num)
                    and self._get_bg_control(bg_num).priority == priority
                ):
                    self._merge_layer(self.bg_scanline[bg_num], bg_num, priority)

            if self.display_control.display_obj:
                self._merge_layer(self.obj_scanline[priority], LayerType.OBJ, priority)

        for x in range(DISPLAY_WIDTH):
            self.back_buffer[DISPLAY_WIDTH * self.vcount + x] = self.scanline[x]

    def _merge_layer(
        self,
        layer_scanline,
        layer_type: LayerType | int,
        priority: int,
    ):
        windowing_enabled = (
            self.display_control.display_window(WindowIndex.WIN_0)
            or self.display_control.display_window(WindowIndex.WIN_1)
            or self.display_control.display_window(WindowIndex.WIN_OBJ)
        )

        blend_mode = (
            self.blend_control.blend_mode
            if self.blend_control.blend_source(layer_type)
            else BlendMode.OFF
        )

        coefficient_source = self.blend_alpha.coefficient_source
        coefficient_target = self.blend_alpha.coefficient_target
        brightness = self.blend_brightness.brightness

        for x in range(DISPLAY_WIDTH):
            pixel_blend_mode = blend_mode

            if windowing_enabled:
                window_control = self._get_window_for_pixel(x)
                if not window_control.display_layer(layer_type):
                    continue
                if not window_control.enable_blending:
                    pixel_blend_mode = BlendMode.OFF

            if layer_type == LayerType.OBJ and self.blend_mask_obj[priority][x]:
                pixel_blend_mode = BlendMode.ALPHA

            colour = layer_scanline[x]
            if colour == TRANSPARENT_COLOUR:
                continue

            if pixel_blend_mode == BlendMode.OFF:
                self.scanline[x] = colour
            elif pixel_blend_mode == BlendMode.ALPHA:
                if self.blend_control.blend_target(self.scanline_top_layer[x]):
                    self.scanline[x] = blend_colours(
                        colour,
                        self.scanline_top_colour[x],
                        coefficient_source,
                        coefficient_target,
                    )
                else:
                    self.scanline[x] = colour
            elif pixel_blend_mode == BlendMode.LIGHTEN:
                self.scanline[x] = colour
            elif pixel_blend_mode == BlendMode.DARKEN:
                self.scanline[x] = colour

            self.scanline_top_colour[x] = colour
            self.scanline_top_layer[x] = layer_type

    def _calc_window_masks(self):
        for window in (WindowIndex.WIN_0, WindowIndex.WIN_1):
            window_mask = self.window_mask_0 if window == WindowIndex.WIN_0 else self.window_mask_1

            h_min = self.window_h_min[int(window)]
            h_max = self.window_h_max[int(window)]
            v_min = self.window_v_min[int(window)]
            v_max = self.window_v_max[int(window)]

            inside_v_bounds = (
                v_min <= self.vcount < v_max
                if v_min <= v_max
                else (self.vcount < v_max or self.vcount >= v_min)
            )

            if not self.display_control.display_window(window) or not inside_v_bounds:
                for x in range(DISPLAY_WIDTH):
                    window_mask[x] = 0
            else:
                if h_min <= h_max:
                    for x in range(DISPLAY_WIDTH):
                        window_mask[x] = h_min <= x < h_max
                else:
                    for x in range(DISPLAY_WIDTH):
                        window_mask[x] = x < h_max or x >= h_min

        for x in range(DISPLAY_WIDTH):
            self.window_mask_obj[x] = 0

    def _get_window_for_pixel(self, x):
        if self.window_mask_0[x]:
            return self.window_control_0
        elif self.window_mask_1[x]:
            return self.window_control_1
        elif self.window_mask_obj[x]:
            return self.window_control_obj
        else:
            return self.window_control_out

    def _get_palette_index(
        self,
        tile_base_address: int,
        tile_index: int,
        pixel_x: int,
        pixel_y: int,
        colour_256: bint,
        palette_num: int,
    ) -> int:
        if colour_256:
            # 256-colour object; one byte per pixel, one palette with 256 colours
            pixel_address = tile_base_address
            pixel_address += tile_index * TILE_SIZE + pixel_y * TILE_WIDTH + pixel_x
            palette_index = self.memory.read_8_vram(pixel_address)
        else:
            # 16-colour object; each byte represents two pixels

            # Lower 4 bits are for the left pixel and upper 4 bits are for the right pixel
            pixel_address = tile_base_address
            pixel_address += (tile_index * TILE_SIZE + pixel_y * TILE_WIDTH + pixel_x) // 2
            palette_index = (self.memory.read_8_vram(pixel_address) >> ((pixel_x % 2) * 4)) & 0xF

            if palette_index == 0:
                return 0

            # Palette num selects one of 16 palettes, each palette having 16 colours
            palette_index += palette_num * 16

        return palette_index

    def _get_bg_control(self, bg_num: int) -> BackgroundControlRegister:
        if bg_num == 0:
            return self.bg_control_0
        elif bg_num == 1:
            return self.bg_control_1
        elif bg_num == 2:
            return self.bg_control_2
        elif bg_num == 3:
            return self.bg_control_3
        # ifndef CYTHON
        else:
            raise ValueError("Invalid bg_num")
        # endif


class Object:
    """
    Corresponds to an OAM entry representing an object (moveable sprite).
    Each entry consists of three 16-bit attributes.
    """

    def __init__(self, attr_0: int, attr_1: int, attr_2: int):
        self.SIZES = [
            [(8, 8), (16, 16), (32, 32), (64, 64)],  # Square
            [(16, 8), (32, 8), (32, 16), (64, 32)],  # Horizontal
            [(8, 16), (8, 32), (16, 32), (32, 64)],  # Vertical
        ]

        self.attr_0 = attr_0
        self.attr_1 = attr_1
        self.attr_2 = attr_2

    @property
    def x(self) -> int:
        return get_bits(self.attr_1, 0, 8)

    @property
    def y(self) -> int:
        return get_bits(self.attr_0, 0, 7)

    @property
    def affine(self) -> bint:
        return get_bit(self.attr_0, 8)

    @property
    def double_size(self) -> bint:
        return get_bit(self.attr_0, 9)

    @property
    def disabled(self) -> bint:
        return get_bit(self.attr_0, 9) & ~self.affine

    @property
    def mode(self) -> ObjectMode | int:
        return get_bits(self.attr_0, 10, 11)

    @property
    def mosaic(self) -> bint:
        return get_bit(self.attr_0, 12)

    @property
    def colour_256(self) -> bint:
        return get_bit(self.attr_0, 13)

    @property
    def shape(self) -> ObjectShape | int:
        return get_bits(self.attr_0, 14, 15)

    @property
    def flip_horizontal(self) -> bint:
        return get_bit(self.attr_1, 12) & ~self.affine

    @property
    def flip_vertical(self) -> bint:
        return get_bit(self.attr_1, 13) & ~self.affine

    @property
    def size(self) -> tuple[int, int]:
        size = get_bits(self.attr_1, 14, 15)
        return self.SIZES[self.shape][size]

    @property
    def tile_index(self) -> int:
        return get_bits(self.attr_2, 0, 9)

    @property
    def priority(self) -> int:
        return get_bits(self.attr_2, 10, 11)

    @property
    def palette_num(self) -> int:
        return get_bits(self.attr_2, 12, 15)


def get_rgb_channels(colour: int) -> tuple[int, int, int]:
    r = colour & 0x1F
    g = (colour >> 5) & 0x1F
    b = (colour >> 10) & 0x1F
    return r, g, b


def blend_colours(
    colour_a: int,
    colour_b: int,
    coefficient_a: int,
    coefficient_b: int,
) -> int:
    r_a, g_a, b_a = get_rgb_channels(colour_a)
    r_b, g_b, b_b = get_rgb_channels(colour_b)
    r = min(31, (r_a * coefficient_a + r_b * coefficient_b) >> 4)
    g = min(31, (g_a * coefficient_a + g_b * coefficient_b) >> 4)
    b = min(31, (b_a * coefficient_a + b_b * coefficient_b) >> 4)
    return r | (g << 5) | (b << 10)



# ifndef CYTHON
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.cpu.cpu import CPU

from pyboy_advance.cpu.constants import CPUMode
from pyboy_advance.memory.constants import MemoryAccess
from pyboy_advance.utils import (
    get_bit,
    get_bits,
    set_bit,
    add_32,
)
# endif


def arm_block_data_transfer(cpu: CPU, instr: int):
    """Execute a Block Data Transfer instruction (LDM, STM)"""

    pre_post_bit = get_bit(instr, 24)
    up_down_bit = get_bit(instr, 23)
    psr_and_force_user_bit = get_bit(instr, 22)
    write_back_bit = get_bit(instr, 21)
    load_store_bit = get_bit(instr, 20)

    base_reg = get_bits(instr, 16, 19)
    reg_list = get_bits(instr, 0, 15)

    # Avoiding usage of sum() here to optimize Cython output
    reg_list_count = 0
    for reg in range(16):
        reg_list_count += get_bit(reg_list, reg)

    # If the reg list is empty, PC is loaded/stored and base is
    # incremented as if all regs were transferred
    if reg_list_count == 0:
        reg_list = set_bit(reg_list, cpu.regs.PC, True)
        reg_list_count = 16

    pc_in_reg_list = get_bit(reg_list, cpu.regs.PC)

    base = cpu.regs.get(base_reg)

    # Find the final address
    if up_down_bit:
        # Up: add offset to base
        final_address = add_32(base, reg_list_count * 4)
    else:
        # Down: subtract offset from base
        final_address = add_32(base, reg_list_count * -4)

        # Move base to the final address and increment from there,
        # instead of decrementing from original base
        base = final_address
        pre_post_bit = not pre_post_bit

    # When S bit (psr_and_force_user_bit) is set and either the instruction is STM
    # or PC is not in the list, this indicates a User Bank Transfer.
    # The reg list refers to registers from the User bank rather than the bank
    # corresponding to the current mode.
    original_mode = cpu.regs.cpsr.mode
    if psr_and_force_user_bit and not (pc_in_reg_list and load_store_bit):
        cpu.switch_mode(CPUMode.USER)

    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.NON_SEQUENTIAL

    first = True
    access_type = MemoryAccess.NON_SEQUENTIAL
    for reg in range(16):
        if get_bit(reg_list, reg):
            if pre_post_bit:  # Pre: add offset before transfer
                base = add_32(base, 4)

            if load_store_bit:  # Load
                # Write back before read
                if first and write_back_bit:
                    first = False
                    cpu.regs.set(base_reg, final_address)

                cpu.regs.set(reg, cpu.memory.read_32(base, access_type))
            else:  # Store
                cpu.memory.write_32(base, cpu.regs.get(reg), access_type)

                # Write back after write
                if first and write_back_bit:
                    first = False
                    cpu.regs.set(base_reg, final_address)

            if not pre_post_bit:  # Post: add offset after transfer
                base = add_32(base, 4)

            access_type = MemoryAccess.SEQUENTIAL

    # When S bit (psr_and_force_user_bit) is set and the instruction is LDM and
    # PC is in the list, this indicates we need to change CPSR=SPSR_<current mode>.
    if pc_in_reg_list and load_store_bit:
        if psr_and_force_user_bit:
            spsr_reg = cpu.regs.spsr.reg
            spsr_mode = cpu.regs.spsr.mode
            cpu.switch_mode(spsr_mode)
            cpu.regs.cpsr.reg = spsr_reg

        # The pipeline also needs to be flushed since PC was changed
        cpu.flush_pipeline()

    # Otherwise, the S bit indicates a User Bank Transfer, so if it's ON then we
    # must have switched into user mode earlier. Hence, we need to switch back
    # into the mode we were in before the transfer.
    elif psr_and_force_user_bit:
        cpu.switch_mode(original_mode)

    if load_store_bit:
        cpu.scheduler.idle(1)



from libc.stdint cimport uint32_t

from pyboy_advance.cpu.cpu cimport CPU
from pyboy_advance.memory.constants cimport MemoryAccess
from pyboy_advance.utils cimport get_bit, get_bits

cdef void arm_single_data_swap(CPU, uint32_t) noexcept



from libc.stdint cimport uint32_t

from pyboy_advance.cpu.cpu cimport CPU
from pyboy_advance.memory.constants cimport MemoryAccess
from pyboy_advance.utils cimport (
    get_bits,
    extend_sign_12,
    extend_sign_8,
    get_bit,
    add_32,
    extend_sign_23,
)

cdef void thumb_unconditional_branch(CPU, uint32_t) noexcept
cdef void thumb_conditional_branch(CPU, uint32_t) noexcept
cdef void thumb_long_branch_with_link(CPU, uint32_t) noexcept



cimport cython

from libc.stdint cimport uint8_t, uint32_t, uint64_t

cdef bint get_bit(uint32_t, int) noexcept

cdef uint32_t get_bits(uint32_t, int, int) noexcept

cdef uint32_t set_bit(uint32_t, int, bint) noexcept

cdef bint sign_32(uint32_t) noexcept

cdef bint sign_24(uint32_t) noexcept

cdef bint sign_23(uint32_t) noexcept

cdef bint sign_16(uint32_t) noexcept

cdef bint sign_12(uint32_t) noexcept

cdef bint sign_9(uint32_t) noexcept

cdef bint sign_8(uint32_t) noexcept

@cython.locals(extension=uint64_t)
cdef uint64_t extend_sign_32(uint32_t) noexcept

@cython.locals(extension=uint32_t)
cdef uint32_t extend_sign_24(uint32_t) noexcept

@cython.locals(extension=uint32_t)
cdef uint32_t extend_sign_23(uint32_t) noexcept

@cython.locals(extension=uint32_t)
cdef uint32_t extend_sign_16(uint32_t) noexcept

@cython.locals(extension=uint32_t)
cdef uint32_t extend_sign_12(uint32_t) noexcept

@cython.locals(extension=uint32_t)
cdef uint32_t extend_sign_9(uint32_t) noexcept

@cython.locals(extension=uint32_t)
cdef uint32_t extend_sign_8(uint32_t) noexcept

@cython.locals(mask=uint32_t)
cdef uint32_t add_32(uint32_t, uint32_t) noexcept

@cython.locals(mask=uint32_t)
cdef uint32_t ror_32(uint32_t, int) noexcept

cdef uint32_t array_read_32(uint8_t[:], int) noexcept

cdef uint32_t array_read_16(uint8_t[:], int) noexcept

cdef void array_write_32(uint8_t[:], int, uint32_t) noexcept

cdef void array_write_16(uint8_t[:], int, uint32_t) noexcept



# ifndef CYTHON
from __future__ import annotations

from pyboy_advance.constants import EventTrigger, Interrupt
from pyboy_advance.interrupt_controller import InterruptController
from pyboy_advance.scheduler import Scheduler, Event
from pyboy_advance.utils import bint, get_bit, get_bits, set_bit
# endif


class Timers:
    def __init__(
        self,
        scheduler: Scheduler,
        interrupt_controller: InterruptController,
    ):
        self.timer_3 = Timer(scheduler, interrupt_controller, Interrupt.TIMER_3, None)
        self.timer_2 = Timer(scheduler, interrupt_controller, Interrupt.TIMER_2, self.timer_3)
        self.timer_1 = Timer(scheduler, interrupt_controller, Interrupt.TIMER_1, self.timer_2)
        self.timer_0 = Timer(scheduler, interrupt_controller, Interrupt.TIMER_0, self.timer_1)


class Timer:
    def __init__(
        self,
        scheduler: Scheduler,
        interrupt_controller: InterruptController,
        interrupt: Interrupt,
        next_timer: Timer | None,
    ):
        self.WRITE_TIMER_REGISTER_DELAY = 1
        self.START_TIMER_DELAY = 1
        self.FREQUENCIES = [1, 64, 256, 1024]

        self.scheduler = scheduler
        self.interrupt_controller = interrupt_controller
        self.interrupt = interrupt
        self.next_timer = next_timer

        self._counter = 0
        self._reload_value = 0
        self._control_reg = 0

        self._pending_reload_value = 0
        self._pending_control_reg = 0

        self._overflow_event: Event | None = None

    @property
    def counter(self) -> int:
        if self.timer_enabled and not self.count_up:
            self._update_counter()
        return self._counter

    @counter.setter
    def counter(self, value: int):
        self._pending_reload_value = value
        self.scheduler.schedule(
            self._write_reload_value,
            self.WRITE_TIMER_REGISTER_DELAY,
            EventTrigger.IMMEDIATELY,
        )

    @property
    def control_reg(self) -> int:
        return self._control_reg

    @control_reg.setter
    def control_reg(self, value: int):
        self._pending_control_reg = value
        self.scheduler.schedule(
            self._write_control_reg,
            self.WRITE_TIMER_REGISTER_DELAY,
            EventTrigger.IMMEDIATELY,
        )

    @property
    def frequency(self) -> int:
        return self.FREQUENCIES[get_bits(self._control_reg, 0, 1)]

    @property
    def count_up(self) -> bint:
        return get_bit(self._control_reg, 2)

    @count_up.setter
    def count_up(self, value: bint):
        set_bit(self._control_reg, 2, value)

    @property
    def irq_enabled(self) -> bint:
        return get_bit(self._control_reg, 6)

    @property
    def timer_enabled(self) -> bint:
        return get_bit(self._control_reg, 7)

    def _write_reload_value(self):
        self._reload_value = self._pending_reload_value

    def _write_control_reg(self):
        old_count_up = self.count_up
        old_timer_enabled = self.timer_enabled

        self._control_reg = self._pending_control_reg

        new_count_up = self.count_up
        new_timer_enabled = self.timer_enabled

        if new_timer_enabled and not old_timer_enabled:
            self._start()
        elif not new_timer_enabled and old_timer_enabled:
            self._stop()
        elif new_count_up and not old_count_up:
            self._stop()
        elif not new_count_up and old_count_up:
            self._start()

    def _start(self):
        if not self.count_up:
            self._counter = self._reload_value
            self._schedule_overflow_event(self.START_TIMER_DELAY)
        elif self._overflow_event:
            self._schedule_overflow_event(0)

    def _stop(self):
        self._update_counter()

        if self._overflow_event:
            self._overflow_event.cancelled = True
            self._overflow_event = None

    def _update_counter(self):
        if self._overflow_event:
            elapsed = (self.scheduler.cycles - self._overflow_event.time) & 0xFFFF
            self._counter = elapsed // self.frequency

    def _schedule_overflow_event(self, start_delay: int):
        delay = (0x10000 - self._counter) * self.frequency + start_delay

        if self._overflow_event:
            self._overflow_event.cancelled = True
            self._overflow_event = None

        self._overflow_event = self.scheduler.schedule(
            self._overflow,
            delay,
            EventTrigger.IMMEDIATELY,
        )

    def _overflow(self):
        self._counter = self._reload_value

        if not self.count_up:
            self._schedule_overflow_event(0)
        elif self._overflow_event:
            self._overflow_event.cancelled = True
            self._overflow_event = None

        if self.irq_enabled:
            self.interrupt_controller.signal(self.interrupt)

        if self.next_timer and self.next_timer.count_up and self.next_timer.timer_enabled:
            self.next_timer._counter += 1
            if self.next_timer._counter == 0x10000:
                self.next_timer._overflow()



from enum import IntEnum

class CPUState(IntEnum):
    ARM = 0
    THUMB = 1

class CPUMode(IntEnum):
    USER = 16
    FIQ = 17
    IRQ = 18
    SWI = 19
    ABORT = 23
    UNDEFINED = 27
    SYSTEM = 31

class Condition(IntEnum):
    EQ = 0
    NE = 1
    HS = 2
    LO = 3
    MI = 4
    PL = 5
    VS = 6
    VC = 7
    HI = 8
    LS = 9
    GE = 10
    LT = 11
    GT = 12
    LE = 13
    AL = 14
    NV = 15

class ShiftType(IntEnum):
    LSL = 0
    LSR = 1
    ASR = 2
    ROR = 3

class ExceptionVector(IntEnum):
    RESET = 0
    UNDEFINED_INSTRUCTION = 4
    SWI = 8
    PREFETCH_ABORT = 12
    DATA_ABORT = 16
    ADDRESS_EXCEEDS_26_BITS = 20
    IRQ = 24
    FIQ = 28

class BankIndex(IntEnum):
    BANK_SYSTEM_USER = 0
    BANK_FIQ = 1
    BANK_IRQ = 2
    BANK_SWI = 3
    BANK_ABORT = 4
    BANK_UNDEFINED = 5
ARM_PC_INCREMENT = 4
THUMB_PC_INCREMENT = 2


def arm_block_data_transfer(cpu: CPU, instr: int):
    """Execute a Block Data Transfer instruction (LDM, STM)"""
    pre_post_bit = get_bit(instr, 24)
    up_down_bit = get_bit(instr, 23)
    psr_and_force_user_bit = get_bit(instr, 22)
    write_back_bit = get_bit(instr, 21)
    load_store_bit = get_bit(instr, 20)
    base_reg = get_bits(instr, 16, 19)
    reg_list = get_bits(instr, 0, 15)
    reg_list_count = 0
    for reg in range(16):
        reg_list_count += get_bit(reg_list, reg)
    if reg_list_count == 0:
        reg_list = set_bit(reg_list, cpu.regs.PC, True)
        reg_list_count = 16
    pc_in_reg_list = get_bit(reg_list, cpu.regs.PC)
    base = cpu.regs.get(base_reg)
    if up_down_bit:
        final_address = add_32(base, reg_list_count * 4)
    else:
        final_address = add_32(base, reg_list_count * -4)
        base = final_address
        pre_post_bit = not pre_post_bit
    original_mode = cpu.regs.cpsr.get_mode()
    if psr_and_force_user_bit and (not (pc_in_reg_list and load_store_bit)):
        cpu.switch_mode(CPUMode.USER)
    cpu.advance_pc_arm()
    cpu.next_fetch_access = MemoryAccess.NON_SEQUENTIAL
    first = True
    access_type = MemoryAccess.NON_SEQUENTIAL
    for reg in range(16):
        if get_bit(reg_list, reg):
            if pre_post_bit:
                base = add_32(base, 4)
            if load_store_bit:
                if first and write_back_bit:
                    first = False
                    cpu.regs.set(base_reg, final_address)
                cpu.regs.set(reg, cpu.memory.read_32(base, access_type))
            else:
                cpu.memory.write_32(base, cpu.regs.get(reg), access_type)
                if first and write_back_bit:
                    first = False
                    cpu.regs.set(base_reg, final_address)
            if not pre_post_bit:
                base = add_32(base, 4)
            access_type = MemoryAccess.SEQUENTIAL
    if pc_in_reg_list and load_store_bit:
        if psr_and_force_user_bit:
            spsr_reg = cpu.regs.spsr.reg
            spsr_mode = cpu.regs.spsr.get_mode()
            cpu.switch_mode(spsr_mode)
            cpu.regs.cpsr.reg = spsr_reg
        cpu.flush_pipeline()
    elif psr_and_force_user_bit:
        cpu.switch_mode(original_mode)
    if load_store_bit:
        cpu.scheduler.idle(1)


# ifndef CYTHON
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pyboy_advance.cpu.cpu import CPU

from pyboy_advance.memory.constants import MemoryAccess
from pyboy_advance.utils import (
    get_bits,
    extend_sign_12,
    extend_sign_8,
    get_bit,
    add_32,
    extend_sign_23,
)
# endif


def thumb_unconditional_branch(cpu: CPU, instr: int):
    """Execute a THUMB.18 instruction (unconditional branch)"""

    offset = extend_sign_12(get_bits(instr, 0, 10) << 1)
    cpu.regs.pc = add_32(cpu.regs.pc, offset)
    cpu.flush_pipeline()


def thumb_conditional_branch(cpu: CPU, instr: int):
    """Execute a THUMB.16 instruction (conditional branch)"""

    condition = get_bits(instr, 8, 11)
    offset = extend_sign_8(get_bits(instr, 0, 7)) * 2

    if cpu.check_condition(condition):
        cpu.regs.pc = add_32(cpu.regs.pc, offset)
        cpu.flush_pipeline()
    else:
        cpu.advance_pc_thumb()
        cpu.next_fetch_access = MemoryAccess.SEQUENTIAL


def thumb_long_branch_with_link(cpu: CPU, instr: int):
    """
    Execute a THUMB.19 instruction (long branch with link).

    Long Branch with Link is a 32-bit instruction split across two 16-bit THUMB opcodes.
    Bit 11 determines whether the opcode is the first one or the second one.
    """

    is_first_opcode = not get_bit(instr, 11)
    if is_first_opcode:
        address_upper_bits = extend_sign_23(get_bits(instr, 0, 10) << 12)
        cpu.regs.lr = add_32(cpu.regs.pc, address_upper_bits)
        cpu.advance_pc_thumb()
        cpu.next_fetch_access = MemoryAccess.SEQUENTIAL
    else:
        address_lower_bits = get_bits(instr, 0, 10) << 1
        pc = add_32(cpu.regs.pc, -2)
        cpu.regs.pc = add_32(cpu.regs.lr, address_lower_bits)
        cpu.regs.lr = pc | 1
        cpu.flush_pipeline()



from libc.stdint cimport uint32_t

from pyboy_advance.constants cimport Interrupt, EventTrigger
from pyboy_advance.memory.memory cimport Memory
from pyboy_advance.memory.constants cimport (
    IOAddress,
    MemoryAccess,
    DMAAddressAdjustment,
    DMAStartTiming,
    DMATransferSize,
)
from pyboy_advance.scheduler cimport Scheduler
from pyboy_advance.utils cimport get_bits, get_bit, set_bit

cdef class DMAControlRegister:
    cdef uint32_t reg

    cdef int get_dst_address_adjustment(self) noexcept
    cdef int get_src_address_adjustment(self) noexcept
    cdef bint get_repeat(self) noexcept
    cdef int get_transfer_size(self) noexcept
    cdef int get_start_timing(self) noexcept
    cdef bint get_irq_when_done(self) noexcept
    cdef bint get_transfer_enabled(self) noexcept
    cdef void set_transfer_enabled(self, bint) noexcept

cdef class DMAChannel:
    cdef uint32_t[4] SRC_MASK
    cdef uint32_t[4] DST_MASK
    cdef uint32_t[4] COUNT_MASK
    cdef int[4] INTERRUPT
    cdef int TRANSFER_DELAY

    cdef int channel_id
    cdef Scheduler scheduler
    cdef Memory memory
    cdef bint fifo
    cdef bint pending
    cdef DMAControlRegister _control_reg
    cdef uint32_t _count
    cdef uint32_t _src
    cdef uint32_t _dst
    cdef uint32_t _internal_count
    cdef uint32_t _internal_src
    cdef uint32_t _internal_dst
    cdef object _event

    cdef uint32_t get_control_reg(self) noexcept
    cdef void set_control_reg(self, uint32_t) noexcept
    cdef uint32_t get_count(self) noexcept
    cdef void set_count(self, uint32_t) noexcept
    cdef uint32_t get_src_address(self) noexcept
    cdef void set_src_address(self, uint32_t) noexcept
    cdef uint32_t get_dst_address(self) noexcept
    cdef void set_dst_address(self, uint32_t) noexcept
    cdef void transfer(self) noexcept
    cdef void activate(self) noexcept

cdef class DMAController:
    cdef DMAChannel channel_0
    cdef DMAChannel channel_1
    cdef DMAChannel channel_2
    cdef DMAChannel channel_3

    cdef bint get_active(self) noexcept
    cdef void perform_transfers(self) noexcept


# ifndef CYTHON
from __future__ import annotations

from pyboy_advance.cpu.thumb.alu import (
    thumb_move_shifted_register,
    thumb_add_subtract,
    thumb_move_compare_add_subtract,
    thumb_alu,
    thumb_high_reg_branch_exchange,
    thumb_add_offset_to_stack_pointer,
    thumb_get_address,
)
from pyboy_advance.cpu.thumb.bdt import (
    thumb_multiple_load_store,
    thumb_push_pop_registers,
)
from pyboy_advance.cpu.thumb.branch import (
    thumb_unconditional_branch,
    thumb_conditional_branch,
    thumb_long_branch_with_link,
)
from pyboy_advance.cpu.thumb.sdt import (
    thumb_pc_relative_load,
    thumb_load_store_sign_extended,
    thumb_load_store_register_offset,
    thumb_load_store_immediate_offset,
    thumb_load_store_halfword,
    thumb_sp_relative_load_store,
)
from pyboy_advance.cpu.thumb.swi import thumb_software_interrupt
from pyboy_advance.cpu.arm.decode import InstrHandler, InstrPattern
# endif


THUMB_PATTERNS: list[InstrPattern] = [
    # ----------+------------+------------+------------------------------------
    #           | Mask       | Value      | Handler
    # ----------+------------+------------+------------------------------------
    InstrPattern(0b1111_1111, 0b1101_1111, thumb_software_interrupt),
    InstrPattern(0b1111_1000, 0b1110_0000, thumb_unconditional_branch),
    InstrPattern(0b1111_0000, 0b1101_0000, thumb_conditional_branch),
    InstrPattern(0b1111_0000, 0b1111_0000, thumb_long_branch_with_link),
    InstrPattern(0b1111_0000, 0b1100_0000, thumb_multiple_load_store),
    InstrPattern(0b1111_0110, 0b1011_0100, thumb_push_pop_registers),
    InstrPattern(0b1111_1111, 0b1011_0000, thumb_add_offset_to_stack_pointer),
    InstrPattern(0b1111_0000, 0b1010_0000, thumb_get_address),
    InstrPattern(0b1111_0000, 0b1001_0000, thumb_sp_relative_load_store),
    InstrPattern(0b1111_0000, 0b1000_0000, thumb_load_store_halfword),
    InstrPattern(0b1110_0000, 0b0110_0000, thumb_load_store_immediate_offset),
    InstrPattern(0b1111_0010, 0b0101_0000, thumb_load_store_register_offset),
    InstrPattern(0b1111_0010, 0b0101_0010, thumb_load_store_sign_extended),
    InstrPattern(0b1111_1000, 0b0100_1000, thumb_pc_relative_load),
    InstrPattern(0b1111_1100, 0b0100_0100, thumb_high_reg_branch_exchange),
    InstrPattern(0b1111_1100, 0b0100_0000, thumb_alu),
    InstrPattern(0b1110_0000, 0b0010_0000, thumb_move_compare_add_subtract),
    InstrPattern(0b1111_1000, 0b0001_1000, thumb_add_subtract),
    InstrPattern(0b1110_0000, 0b0000_0000, thumb_move_shifted_register),
]


# ifndef CYTHON
THUMB_LUT: list[InstrHandler | None] = [None] * (2**8)
# endif


def fill_thumb_lut():
    for opcode in range(2**8):
        for pattern in THUMB_PATTERNS:
            if (opcode & pattern.mask) == pattern.value:
                THUMB_LUT[opcode] = pattern.handler
                break


fill_thumb_lut()


def thumb_decoder(instr: int) -> InstrHandler:
    instruction_handler = THUMB_LUT[instr >> 8]

    # ifndef CYTHON
    if not instruction_handler:
        raise ValueError(f"Unknown THUMB instruction: {instr:016b}")
    # endif

    return instruction_handler
